#!/usr/bin/env python
"""Tests for _data_format.py"""

import logging

import numpy as np
import pandas as pd
import pytest
import xarray as xr

import primap2
from primap2._selection import translations_from_dims

from .utils import assert_ds_aligned_equal


class TestToNetCDF:
    def test_io_roundtrip(self, any_ds: xr.Dataset, caplog, tmp_path):
        ds = any_ds
        attrs_before = ds.attrs.copy()
        caplog.set_level(logging.INFO)
        ds.pr.to_netcdf(tmp_path / "temp.nc")
        nds = primap2.open_dataset(tmp_path / "temp.nc")
        nds.pr.ensure_valid()
        assert not caplog.records
        xr.testing.assert_identical(ds, nds)
        assert_ds_aligned_equal(ds, nds)
        assert attrs_before == ds.attrs
        assert attrs_before == nds.attrs


class TestEnsureValid:
    def test_something_else_entirely(self, caplog):
        with pytest.raises(ValueError, match=r"ds is not an xr.Dataset"):
            something_else = primap2._data_format.DatasetDataFormatAccessor(ds="asdf")
            something_else.ensure_valid()
        assert "ERROR" in caplog.text
        assert "object is not an xarray Dataset." in caplog.text

    def test_valid_ds_pass(self, any_ds, caplog):
        caplog.set_level(logging.INFO)
        any_ds.pr.ensure_valid()
        assert not caplog.records

    def test_time_dimension_for_metadata(self, opulent_processing_ds, caplog):
        opulent_processing_ds["Processing of CO2"] = opulent_processing_ds[
            "Processing of CO2"
        ].expand_dims(dim={"time": np.array(["2020", "2021"], dtype=np.datetime64)})
        with pytest.raises(
            ValueError, match=r"contains metadata, but carries 'time' dimension"
        ):
            opulent_processing_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert "'Processing of CO2' is a metadata variable, but 'time' is a dimension."

    def test_metadata_missing_attr(self, opulent_processing_ds, caplog):
        del opulent_processing_ds["Processing of CO2"].attrs["described_variable"]
        with pytest.raises(
            ValueError,
            match=r"'described_variable' attr missing for 'Processing of CO2'",
        ):
            opulent_processing_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text

    def test_metadata_wrong_attr(self, opulent_processing_ds, caplog):
        opulent_processing_ds["Processing of CO2"].attrs["described_variable"] = "CH4"
        with pytest.raises(
            ValueError,
            match=r"variable name 'Processing of CO2' inconsistent with "
            r"described_variable 'CH4'",
        ):
            opulent_processing_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text

    def test_required_dimension_missing(self, caplog):
        ds = xr.Dataset(
            {
                "area (ISO3)": ["a"],
                "time": pd.date_range("2000-01-01", "2020-01-01", freq="YS"),
            },
            attrs={"area": "area (ISO3)"},
        ).pr.quantify()

        with pytest.raises(ValueError, match=r"'source' not in dims"):
            ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert "'source' not found in dims, but is required." in caplog.text

    def test_required_dimension_missing_var(self, minimal_ds, caplog):
        minimal_ds["CO2"] = minimal_ds["CO2"].squeeze("source", drop=True)

        with pytest.raises(ValueError, match=r"'source' not in dims"):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert (
            "'source' not found in dims for variable 'CO2', but is required"
            in caplog.text
        )

    def test_required_coordinate_missing(self, minimal_ds, caplog):
        del minimal_ds["source"]
        with pytest.raises(ValueError, match=r"dim 'source' has no coord"):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert "No coord found for dimension 'source'." in caplog.text

    def test_dimension_metadata_missing(self, minimal_ds, caplog):
        del minimal_ds.attrs["area"]
        with pytest.raises(ValueError, match=r"'area' not in attrs"):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert (
            "'area' not found in attrs, required dimension is therefore undefined."
            in caplog.text
        )

    def test_dimension_metadata_wrong(self, minimal_ds, caplog):
        minimal_ds.attrs["area"] = "asdf"
        with pytest.raises(ValueError, match=r"'area' dimension not in dims"):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert (
            "'asdf' defined as 'area' dimension, but not found in dims." in caplog.text
        )

    def test_wrong_provenance_value(self, opulent_ds, caplog):
        opulent_ds["provenance"] = ["asdf"]
        with pytest.raises(ValueError, match=r"Invalid provenance: \{'asdf'\}"):
            opulent_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert "provenance contains invalid values: {'asdf'}" in caplog.text

    def test_additional_dimension(self, minimal_ds: xr.Dataset, caplog):
        ds = minimal_ds.expand_dims({"addl_dim": ["a", "b", "c"]})  # type: ignore
        ds.pr.ensure_valid()
        assert "WARNING" in caplog.text
        assert (
            "Dimension(s) {'addl_dim'} unknown, likely a typo or missing in sec_cats."
            in caplog.text
        )

    def test_wrong_dimension_key(self, minimal_ds, caplog):
        ds = minimal_ds.rename_dims({"area (ISO3)": "asdf"})
        ds.attrs["area"] = "asdf"
        with pytest.raises(
            ValueError, match=r"'asdf' not in the format 'dim \(category_set\)'"
        ):
            ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert "'asdf' not in the format 'dim (category_set)'." in caplog.text

    def test_missing_sec_cat(self, minimal_ds, caplog):
        minimal_ds.attrs["sec_cats"] = ["missing"]
        with pytest.raises(
            ValueError, match="Secondary category 'missing' not in dims"
        ):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert (
            "Secondary category 'missing' defined, but not found in dims:"
            in caplog.text
        )

    def test_missing_optional_dim(self, minimal_ds, caplog):
        minimal_ds.attrs["scen"] = "missing"
        with pytest.raises(ValueError, match="'scen' not in dims"):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert "'missing' defined as scen, but not found in dims." in caplog.text

    def test_sec_cat_without_primary_cat(self, minimal_ds, caplog):
        ds = minimal_ds.expand_dims({"something (cset)": ["a", "b", "c"]})
        ds.attrs["sec_cats"] = ["something (cset)"]
        ds.pr.ensure_valid()
        assert "WARNING" in caplog.text
        assert (
            "Secondary category defined, but no primary category defined, weird."
            in caplog.text
        )

    def test_additional_coordinate_space(self, opulent_ds: xr.Dataset, caplog):
        ds = opulent_ds.rename({"category_names": "category names"})
        with pytest.raises(
            ValueError, match=r"Coord key 'category names' contains a space"
        ):
            ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert (
            "Additional coordinate name 'category names' contains a space,"
            " replace it with an underscore." in caplog.text
        )

    def test_missing_entity(self, minimal_ds, caplog):
        del minimal_ds["CO2"].attrs["entity"]
        with pytest.raises(ValueError, match="entity missing for 'CO2'"):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert "'CO2' has no entity declared in attributes." in caplog.text

    def test_missing_gwp_context(self, minimal_ds, caplog):
        del minimal_ds["SF6 (SARGWP100)"].attrs["gwp_context"]
        minimal_ds.pr.ensure_valid()
        assert "WARNING" in caplog.text
        assert (
            "'SF6 (SARGWP100)' has the dimension [CO2 * mass / time], but is not CO2. "
            "gwp_context missing?" in caplog.text
        )

    def test_wrong_units(self, minimal_ds, caplog):
        deq = minimal_ds.pint.dequantify()
        deq["CO2"].attrs["units"] = "kg CO2"
        req = deq.pr.quantify()
        req.pr.ensure_valid()
        assert "WARNING" in caplog.text
        assert (
            "'CO2' has a unit of CO2 * kilogram, which is not compatible with an"
            " emission rate." in caplog.text
        )

    def test_unquantified(self, minimal_ds, caplog):
        deq = minimal_ds.pint.dequantify()
        deq.pr.ensure_valid()
        assert not caplog.records

    def test_multi_units(self, minimal_ds, caplog):
        minimal_ds["CO2"].attrs["units"] = "kg CO2 / year"
        with pytest.raises(ValueError, match="data already has units"):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert (
            "'units' in variable attrs, but data is quantified already." in caplog.text
        )

    def test_invalid_units(self, minimal_ds, caplog):
        deq = minimal_ds.pint.dequantify()
        deq["CO2"].attrs["units"] = "i_am_not_a_unit"
        with pytest.raises(ValueError, match="Cannot parse units"):
            deq.pr.ensure_valid()

    def test_invalid_gwp_context(self, minimal_ds, caplog):
        minimal_ds["SF6 (SARGWP100)"].attrs["gwp_context"] = "i_am_not_a_gwp_context"
        with pytest.raises(
            ValueError,
            match=r"Invalid gwp_context 'i_am_not_a_gwp_context' for "
            r"'SF6 \(SARGWP100\)'",
        ):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert (
            "gwp_context 'i_am_not_a_gwp_context' for 'SF6 (SARGWP100)' is not valid."
            in caplog.text
        )

    def test_extraneous_gwp_context(self, minimal_ds, caplog):
        minimal_ds["SF6"].attrs["gwp_context"] = "SARGWP100"
        with pytest.raises(
            ValueError, match=r"SF6 has wrong dimensionality for gwp_context."
        ):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert (
            "'SF6' is a global warming potential, but the dimension is not "
            "[CO2 * mass / time]." in caplog.text
        )

    def test_missing_unit(self, minimal_ds, caplog):
        minimal_ds["CO2"] = (minimal_ds["CO2"].dims, minimal_ds["CO2"].pint.magnitude)
        minimal_ds["CO2"].attrs["entity"] = "CO2"
        with pytest.raises(ValueError, match=r"units missing for 'CO2'"):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert "'CO2' is numerical (float) data, but has no units." in caplog.text

    def test_weird_variable_name(self, minimal_ds, caplog):
        caplog.set_level(logging.INFO)
        minimal_ds["weird_name"] = minimal_ds["CO2"]
        minimal_ds["weird_name"].attrs["entity"] = "CO2"
        minimal_ds.pr.ensure_valid()
        assert "INFO" in caplog.text
        assert "The name 'weird_name' is not in standard format 'CO2'." in caplog.text

    def test_missing_gwp_in_variable_name(self, minimal_ds, caplog):
        minimal_ds["SF6_gwp"] = minimal_ds["SF6 (SARGWP100)"]
        minimal_ds.pr.ensure_valid()
        assert "WARNING" in caplog.text
        assert (
            "'SF6_gwp' has a gwp_context in attrs, but not in its name." in caplog.text
        )

    def test_weird_contact(self, minimal_ds, caplog):
        caplog.set_level(logging.INFO)
        minimal_ds.attrs["contact"] = "i_am_not_an_email"
        minimal_ds.pr.ensure_valid()
        assert "INFO" in caplog.text
        assert "Contact information is not an email address" in caplog.text

    def test_weird_references(self, minimal_ds, caplog):
        caplog.set_level(logging.INFO)
        minimal_ds.attrs["references"] = "i_am_not_a_doi"
        minimal_ds.pr.ensure_valid()
        assert "INFO" in caplog.text
        assert "Reference information is not a DOI" in caplog.text

    def test_nonstandard_attribute(self, minimal_ds, caplog):
        minimal_ds.attrs["i_am_not_standard"] = ""
        minimal_ds.pr.ensure_valid()
        assert "WARNING" in caplog.text
        assert "Unknown metadata in attrs: {'i_am_not_standard'}, typo?" in caplog.text

    def test_publication_date_not_date(self, minimal_ds, caplog):
        minimal_ds.attrs["publication_date"] = "2020-12-31"
        with pytest.raises(ValueError, match="not a date"):
            minimal_ds.pr.ensure_valid()
        assert "ERROR" in caplog.text
        assert "not a datetime.date object" in caplog.text


class TestToInterchangeFormat:
    def test_error_additional_coordinate_dimensions(self, opulent_ds, caplog):
        coords = {
            "time": opulent_ds["time"],
            "area (ISO3)": opulent_ds["area (ISO3)"],
        }
        ds = opulent_ds.assign_coords(
            {
                "addl_coord_2d": xr.DataArray(
                    data=np.zeros([len(x) for x in coords.values()]),
                    dims=coords.keys(),
                    coords=coords,
                )
            }
        )
        with pytest.raises(ValueError, match="has more than one dimension"):
            ds.pr.to_interchange_format()

        assert "ERROR" in caplog.text
        assert (
            "Additional coordinate 'addl_coord_2d' has more than one dimension, which"
            " is not supported." in caplog.text
        )


def test_remove_processing_info(opulent_processing_ds):
    result = opulent_processing_ds.pr.remove_processing_info()
    assert all(not x.startswith("Processing of") for x in result)


def test_remove_processing_info_nothing_to_do(opulent_ds):
    result = opulent_ds.pr.remove_processing_info()
    xr.testing.assert_identical(opulent_ds, result)


def test_has_processing_info_not(opulent_ds):
    assert not opulent_ds.pr.has_processing_info()


def test_has_processing_info(opulent_processing_ds):
    assert opulent_processing_ds.pr.has_processing_info()


def test_expand_dims(minimal_ds):
    result_ds = minimal_ds.pr.expand_dims(
        dim="new_dim",
        coord_value="new_value",
        terminology="new_terminology",
    )

    new_dim = "new_dim (new_terminology)"

    assert new_dim in result_ds.coords
    assert "new_value" in result_ds.coords[new_dim].values

    translations = translations_from_dims([new_dim])
    shortest_key = min(list(translations.keys()), key=len)

    assert result_ds.attrs[shortest_key] == translations[shortest_key]

    # without terminology
    result_ds = minimal_ds.pr.expand_dims(
        dim="new_dim",
        coord_value="new_value",
    )

    new_dim = "new_dim"

    assert new_dim in result_ds.coords
    assert "new_value" in result_ds.coords[new_dim].values



import datetime
import itertools
import re
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import IO, Any

import numpy as np
import pandas as pd
import pint
from loguru import logger

from .. import _selection
from .._units import ureg
from . import _conversion
from ._interchange_format import (
    INTERCHANGE_FORMAT_COLUMN_ORDER,
    INTERCHANGE_FORMAT_MANDATORY_COLUMNS,
    INTERCHANGE_FORMAT_OPTIONAL_COLUMNS,
)

SEC_CATS_PREFIX = "sec_cats__"

BASKET_UNITS = [
    "KYOTOGHG",
    "FGASES",
    "HFCS",
    "PFCS",
    "OTHERHFCS",
    "OTHERPFCS",
]


def convert_long_dataframe_if(
    data_long: pd.DataFrame,
    *,
    coords_cols: dict[str, str],
    add_coords_cols: None | dict[str, list[str]] = None,
    coords_defaults: None | dict[str, Any] = None,
    coords_terminologies: dict[str, str],
    coords_value_mapping: None | dict[str, Any] = None,
    coords_value_filling: None | dict[str, dict[str, dict]] = None,
    filter_keep: None | dict[str, dict[str, Any]] = None,
    filter_remove: None | dict[str, dict[str, Any]] = None,
    meta_data: None | dict[str, Any] = None,
    time_format: str = "%Y-%m-%d",
    convert_str: bool | dict[str, float] = True,
    copy_df: bool = True,
) -> pd.DataFrame:
    """convert a DataFrame in long (tidy) format into the PRIMAP2 interchange format.

    Columns can be renamed or filled with default values to match the PRIMAP2
    structure. Where we refer to "dimensions" in the parameter description below we
    mean the basic dimension names without the added terminology (e.g. "area" not
    "area (ISO3)"). The terminology information will be added by this function. You can
    not use the short dimension names in the attributes (e.g. "cat" instead of
    "category").

    Parameters
    ----------
    data_long: str, pd.DataFrame
        Long format DataFrame file which will be converted.

    coords_cols : dict
        Dict where the keys are column names in the files to be read and the value is
        the dimension in PRIMAP2. To specify the data column containing the observable,
        use the "data" key. For secondary categories use a ``sec_cats__`` prefix.

    add_coords_cols : dict, optional
        Dict where the keys are PRIMAP2 additional coordinate names and the values are
        lists with two elements where the first is the column in the dataframe to be
        converted and the second is the primap2 dimension for the coordinate (e.g.
        ``category`` for a ``category_name`` coordinate).

    coords_defaults : dict, optional
        Dict for default values of coordinates / dimensions not given in the csv files.
        The keys are the dimension names and the values are the values for
        the dimensions. For secondary categories use a ``sec_cats__`` prefix.

    coords_terminologies : dict
        Dict defining the terminologies used for the different coordinates (e.g. ISO3
        for area). Only possible coordinates here are: area, category, scenario,
        entity, and secondary categories. For secondary categories use a ``sec_cats__``
        prefix. All entries different from "area", "category", "scenario", "entity",
        and ``sec_cats__<name>`` will raise a ValueError.

    coords_value_mapping : dict, optional
        A dict with primap2 dimension names as keys. Values are dicts with input values
        as keys and output values as values. A standard use case is to map gas names
        from input data to the standardized names used in primap2.
        Alternatively a value can also be a function which transforms one CSV metadata
        value into the new metadata value.
        A third possibility is to give a string as a value, which defines a rule for
        translating metadata values. For the "category", "entity", and "unit" columns,
        the rule "PRIMAP1" is available, which translates from PRIMAP1 metadata to
        PRIMAP2 metadata.

    coords_value_filling : dict, optional
        A dict with primap2 dimension names as keys. These are the target columns where
        values will be filled (or replaced). Vales are dicts with primap2 dimension
        names as keys. These are the source columns. The values are dicts with source
        value - target value mappings.
        The value filling can do everything that the value mapping can, but while
        mapping can only replace values within a column using information from that
        column, the filing function can also fill or replace data based on values from
        a different column.
        This can be used to e.g. fill missing category codes based on category names or
        to replace category codes which do not meet the terminology using the category
        names.

    filter_keep : dict, optional
        Dict defining filters of data to keep. Filtering is done before metadata
        mapping, so use original metadata values to define the filter. Column names are
        as in the csv file. Each entry in the dict defines an individual filter.
        The names of the filters have no relevance. Default: keep all data.

    filter_remove : dict, optional
        Dict defining filters of data to remove. Filtering is done before metadata
        mapping, so use original metadata values to define the filter. Column names are
        as in the csv file. Each entry in the dict defines an individual filter.
        The names of the filters have no relevance.

    meta_data : dict, optional
        Meta data for the whole dataset. Will end up in the dataset-wide attrs. Allowed
        keys are "references", "rights", "contact", "title", "comment", "institution",
        and "history". Documentation about the format and meaning of the meta data can
        be found in the
        `data format documentation <https://primap2.readthedocs.io/en/stable/\
data_format_details.html#dataset-attributes>`_.

    time_format : str, optional (default: "%Y-%m-%d")
        strftime style format used to format the time information for the data columns
        in the interchange format.
        Default: "%F", i.e. the ISO 8601 date format.

    convert_str : bool or dict, optional (default: True)
        If set to false, string values in the data columns will be kept.
        If set to true they will be converted to np.nan or 0 following default rules.
        If a dict is given mapping will be as given in the dict for values present in
        the dict and default as in parse_code for all other values

    copy_df : bool, optional (default: True)
        If set to true, a copy of the input DataFrame is made to keep the input as is.
        This negatively impacts speed. If set to false the input DataFrame will be
        altered but performance will be better

    Returns
    -------
    obj: pd.DataFrame
        pandas DataFrame with the read data

    Examples
    --------
    *Example for meta_mapping*::

        meta_mapping = {
            'pyCPA_col_1': {'col_1_value_1_in': 'col_1_value_1_out',
                            'col_1_value_2_in': 'col_1_value_2_out',
                            },
            'pyCPA_col_2': {'col_2_value_1_in': 'col_2_value_1_out',
                            'col_2_value_2_in': 'col_2_value_2_out',
                            },
        }

    *Example for filter_keep*::

        filter_keep = {
            'f_1': {'variable': ['CO2', 'CH4'], 'region': 'USA'},
            'f_2': {'variable': 'N2O'}
        }

    This example filter keeps all CO2 and CH4 data for the USA and N2O data for all
    countries

    *Example for filter_remove*::

        filter_remove = {
            'f_1': {'scenario': 'HISTORY'},
        }

    This filter removes all data with 'HISTORY' as scenario

    """
    # Check and prepare arguments
    if coords_defaults is None:
        coords_defaults = {}
    if add_coords_cols is None:
        add_coords_cols = {}
    if meta_data is None:
        attrs = {}
    else:
        attrs = meta_data.copy()

    check_mandatory_dimensions(coords_cols, coords_defaults)
    check_overlapping_specifications(coords_cols, coords_defaults)
    if add_coords_cols:
        check_overlapping_specifications_add_cols(coords_cols, add_coords_cols)

    # if desired make a copy to keep input dataframe unchanged
    if copy_df:
        data_copy = data_long.copy(deep=True)
    else:
        data_copy = data_long

    filter_data(data_copy, filter_keep, filter_remove)

    add_dimensions_from_defaults(
        data_copy, coords_defaults, additional_allowed_coords=["time"]
    )

    naming_attrs = rename_columns(
        data_copy, coords_cols, add_coords_cols, coords_defaults, coords_terminologies
    )
    attrs.update(naming_attrs)

    if convert_str:
        # get data columns (just one as we have long format)
        data_cols = ["data"]
        # find all string values
        str_values = find_str_values_in_data(data_copy, data_cols)
        # create replacement dict
        str_repl_dict = create_str_replacement_dict(str_values, convert_str)
        replace_values(data_copy, data_cols, str_repl_dict)

    additional_coordinates = additional_coordinate_metadata(
        add_coords_cols, coords_cols, coords_terminologies
    )

    if coords_value_mapping is not None:
        map_metadata(data_copy, attrs=attrs, meta_mapping=coords_value_mapping)

    if coords_value_filling is not None:
        data_copy = fill_from_other_col(
            data_copy, attrs=attrs, coords_value_filling=coords_value_filling
        )

    coords = list(set(data_copy.columns.values) - {"data"})

    harmonize_units(data_copy, dimensions=coords, attrs=attrs)

    data_copy["time"] = pd.to_datetime(data_copy["time"], format=time_format)

    data, coords = long_to_wide(data_copy, time_format=time_format)

    data, coords = sort_columns_and_rows(data, dimensions=coords)
    dims = coords.copy()
    for add_coord in add_coords_cols.keys():
        dims.remove(add_coord)

    data.attrs = interchange_format_attrs_dict(
        xr_attrs=attrs,
        time_format=time_format,
        dimensions=dims,
        additional_coordinates=additional_coordinates,
    )

    return data


def read_long_csv_file_if(
    filepath_or_buffer: str | Path | IO,
    *,
    coords_cols: dict[str, str],
    add_coords_cols: None | dict[str, list[str]] = None,
    coords_defaults: None | dict[str, Any] = None,
    coords_terminologies: dict[str, str],
    coords_value_mapping: None | dict[str, Any] = None,
    coords_value_filling: None | dict[str, dict[str, dict]] = None,
    filter_keep: None | dict[str, dict[str, Any]] = None,
    filter_remove: None | dict[str, dict[str, Any]] = None,
    meta_data: None | dict[str, Any] = None,
    time_format: str = "%Y-%m-%d",
    convert_str: bool | dict[str, float] = True,
) -> pd.DataFrame:
    """Read a CSV file in long (tidy) format into the PRIMAP2 interchange format.

    Columns can be renamed or filled with default values to match the PRIMAP2
    structure. Where we refer to "dimensions" in the parameter description below we
    mean the basic dimension names without the added terminology (e.g. "area" not
    "area (ISO3)"). The terminology information will be added by this function. You can
    not use the short dimension names in the attributes (e.g. "cat" instead of
    "category").

    Parameters
    ----------
    filepath_or_buffer: str, pathlib.Path, or file-like
        Long CSV file which will be read.

    coords_cols : dict
        Dict where the keys are column names in the files to be read and the value is
        the dimension in PRIMAP2. To specify the data column containing the observable,
        use the "data" key. For secondary categories use a ``sec_cats__`` prefix.

    add_coords_cols : dict, optional
        Dict where the keys are PRIMAP2 additional coordinate names and the values are
        lists with two elements where the first is the column in the csv file to be
        read and the second is the primap2 dimension for the coordinate (e.g.
        ``category`` for a ``category_name`` coordinate).

    coords_defaults : dict, optional
        Dict for default values of coordinates / dimensions not given in the csv files.
        The keys are the dimension names and the values are the values for
        the dimensions. For secondary categories use a ``sec_cats__`` prefix.

    coords_terminologies : dict
        Dict defining the terminologies used for the different coordinates (e.g. ISO3
        for area). Only possible coordinates here are: area, category, scenario,
        entity, and secondary categories. For secondary categories use a ``sec_cats__``
        prefix. All entries different from "area", "category", "scenario", "entity",
        and ``sec_cats__<name>`` will raise a ValueError.

    coords_value_mapping : dict, optional
        A dict with primap2 dimension names as keys. Values are dicts with input values
        as keys and output values as values. A standard use case is to map gas names
        from input data to the standardized names used in primap2.
        Alternatively a value can also be a function which transforms one CSV metadata
        value into the new metadata value.
        A third possibility is to give a string as a value, which defines a rule for
        translating metadata values. For the "category", "entity", and "unit" columns,
        the rule "PRIMAP1" is available, which translates from PRIMAP1 metadata to
        PRIMAP2 metadata.

    coords_value_filling : dict, optional
        A dict with primap2 dimension names as keys. These are the target columns where
        values will be filled (or replaced). Vales are dicts with primap2 dimension
        names as keys. These are the source columns. The values are dicts with source
        value - target value mappings.
        The value filling can do everything that the value mapping can, but while
        mapping can only replace values within a column using information from that
        column, the filing function can also fill or replace data based on values from
        a different column.
        This can be used to e.g. fill missing category codes based on category names or
        to replace category codes which do not meet the terminology using the category
        names.

    filter_keep : dict, optional
        Dict defining filters of data to keep. Filtering is done before metadata
        mapping, so use original metadata values to define the filter. Column names are
        as in the csv file. Each entry in the dict defines an individual filter.
        The names of the filters have no relevance. Default: keep all data.

    filter_remove : dict, optional
        Dict defining filters of data to remove. Filtering is done before metadata
        mapping, so use original metadata values to define the filter. Column names are
        as in the csv file. Each entry in the dict defines an individual filter.
        The names of the filters have no relevance.

    meta_data : dict, optional
        Meta data for the whole dataset. Will end up in the dataset-wide attrs. Allowed
        keys are "references", "rights", "contact", "title", "comment", "institution",
        and "history". Documentation about the format and meaning of the meta data can
        be found in the
        `data format documentation <https://primap2.readthedocs.io/en/stable/\
data_format_details.html#dataset-attributes>`_.

    time_format : str, optional
        strftime style format used to format the time information for the data columns
        in the interchange format.
        Default: "%F", i.e. the ISO 8601 date format.

    convert_str : bool or dict, optional (default: True)
        If set to false, string values in the data columns will be kept.
        If set to true they will be converted to np.nan or 0 following default rules.
        If a dict is given mapping will be as given in the dict for values present in
        the dict and default as in parse_code for all other values

    Returns
    -------
    obj: pd.DataFrame
        pandas DataFrame with the read data

    Examples
    --------
    *Example for meta_mapping*::

        meta_mapping = {
            'pyCPA_col_1': {'col_1_value_1_in': 'col_1_value_1_out',
                            'col_1_value_2_in': 'col_1_value_2_out',
                            },
            'pyCPA_col_2': {'col_2_value_1_in': 'col_2_value_1_out',
                            'col_2_value_2_in': 'col_2_value_2_out',
                            },
        }

    *Example for filter_keep*::

        filter_keep = {
            'f_1': {'variable': ['CO2', 'CH4'], 'region': 'USA'},
            'f_2': {'variable': 'N2O'}
        }

    This example filter keeps all CO2 and CH4 data for the USA and N2O data for all
    countries

    *Example for filter_remove*::

        filter_remove = {
            'f_1': {'scenario': 'HISTORY'},
        }

    This filter removes all data with 'HISTORY' as scenario

    """
    check_mandatory_dimensions(coords_cols, coords_defaults)
    check_overlapping_specifications(coords_cols, coords_defaults)
    if add_coords_cols:
        check_overlapping_specifications_add_cols(coords_cols, add_coords_cols)

    data_long = read_long_csv(filepath_or_buffer, coords_cols, add_coords_cols)

    return convert_long_dataframe_if(
        data_long=data_long,
        coords_cols=coords_cols,
        add_coords_cols=add_coords_cols,
        coords_defaults=coords_defaults,
        coords_terminologies=coords_terminologies,
        coords_value_mapping=coords_value_mapping,
        coords_value_filling=coords_value_filling,
        filter_keep=filter_keep,
        filter_remove=filter_remove,
        meta_data=meta_data,
        time_format=time_format,
        convert_str=convert_str,
        copy_df=False,
    )


def long_to_wide(
    data_long: pd.DataFrame, *, time_format: str
) -> tuple[pd.DataFrame, list[str]]:
    data_long["time"] = data_long["time"].dt.strftime(time_format)

    coords = list(set(data_long.columns.values) - {"data", "time"})

    # unit is neither a coordinate nor a data column, so has to be handled separately
    unit = data_long[coords].drop_duplicates()
    coords.remove("unit")
    unit.index = pd.MultiIndex.from_frame(unit[coords])

    series = data_long["data"]
    series.index = pd.MultiIndex.from_frame(data_long[[*coords, "time"]])
    data = series.unstack("time")

    data["unit"] = unit["unit"]

    data.reset_index(inplace=True)
    data.columns.name = None

    return data, [*coords, "unit"]


def convert_wide_dataframe_if(
    data_wide: pd.DataFrame,
    *,
    coords_cols: dict[str, str],
    add_coords_cols: None | dict[str, list[str]] = None,
    coords_defaults: None | dict[str, Any] = None,
    coords_terminologies: dict[str, str],
    coords_value_mapping: None | dict[str, Any] = None,
    coords_value_filling: None | dict[str, dict[str, dict]] = None,
    filter_keep: None | dict[str, dict[str, Any]] = None,
    filter_remove: None | dict[str, dict[str, Any]] = None,
    meta_data: None | dict[str, Any] = None,
    time_format: str = "%Y",
    time_cols: None | list = None,
    convert_str: bool | dict[str, float] = True,
    copy_df: bool = False,
) -> pd.DataFrame:
    """
    Convert a DataFrame in wide format into the PRIMAP2 interchange format.

    Columns can be renamed or filled with default values to match the PRIMAP2
    structure. Where we refer to "dimensions" in the parameter description below we
    mean the basic dimension names without the added terminology (e.g. "area" not
    "area (ISO3)"). The terminology information will be added by this function. You can
    not use the short dimension names in the attributes (e.g. "cat" instead of
    "category").

    TODO: Currently duplicate data points will not be detected.

    TODO: enable filtering through query strings

    TODO: enable specification of the entity terminology

    Parameters
    ----------
    data_wide: pd.DataFrame
        Wide DataFrame which will be converted.

    coords_cols : dict
        Dict where the keys are PRIMAP2 dimension names and the values are column
        names in the dataframe to be converted.
        For secondary categories use a ``sec_cats__`` prefix.

    add_coords_cols : dict, optional
        Dict where the keys are PRIMAP2 additional coordinate names and the values are
        lists with two elements where the first is the column in the dataframe to be
        converted and the second is the primap2 dimension for the coordinate (e.g.
        ``category`` for a ``category_name`` coordinate.

    coords_defaults : dict, optional
        Dict for default values of coordinates / dimensions not given in the dataframe.
        The keys are the dimension names and the values are the values for
        the dimensions. For secondary categories use a ``sec_cats__`` prefix.

    coords_terminologies : dict
        Dict defining the terminologies used for the different coordinates (e.g. ISO3
        for area). Only possible coordinates here are: area, category, scenario,
        entity, and secondary categories. For secondary categories use a ``sec_cats__``
        prefix. All entries different from "area", "category", "scenario", "entity",
        and ``sec_cats__<name>`` will raise a ValueError.

    coords_value_mapping : dict, optional
        A dict with primap2 dimension names as keys. Values are dicts with input values
        as keys and output values as values. A standard use case is to map gas names
        from input data to the standardized names used in primap2.
        Alternatively a value can also be a function which transforms one CSV metadata
        value into the new metadata value.
        A third possibility is to give a string as a value, which defines a rule for
        translating metadata values. The only defined rule at the moment is "PRIMAP1"
        which can be used for the "category", "entity", and "unit" columns to translate
        from PRIMAP1 metadata to PRIMAP2 metadata.

    coords_value_filling : dict, optional
        A dict with primap2 dimension names as keys. These are the target columns where
        values will be filled (or replaced). Vales are dicts with primap2 dimension
        names as keys. These are the source columns. The values are dicts with source
        value - target value mappings.
        The value filling can do everything that the value mapping can, but while
        mapping can only replace values within a column using information from that
        column, the filing function can also fill or replace data based on values from
        a different column.
        This can be used to e.g. fill missing category codes based on category names or
        to replace category codes which do not meet the terminology using the category
        names.

    filter_keep : dict, optional
        Dict defining filters of data to keep. Filtering is done before metadata
        mapping, so use original metadata values to define the filter. Column names are
        as in the csv file. Each entry in the dict defines an individual filter.
        The names of the filters have no relevance. Default: keep all data.

    filter_remove : dict, optional
        Dict defining filters of data to remove. Filtering is done before metadata
        mapping, so use original metadata values to define the filter. Column names are
        as in the csv file. Each entry in the dict defines an individual filter.
        The names of the filters have no relevance.

    meta_data : dict, optional
        Meta data for the whole dataset. Will end up in the dataset-wide attrs. Allowed
        keys are "references", "rights", "contact", "title", "comment", "institution",
        and "history". Documentation about the format and meaning of the meta data can
        be found in the
        `data format documentation <https://primap2.readthedocs.io/en/stable/\
data_format_details.html#dataset-attributes>`_.

    time_format : str
        str with strftime style format used to parse the time information for
        the data columns.
        Default: "%Y", which will match years.

    time_cols : list, optional
        List of column names which contain the data for each time point. If not given
        cols will be inferred using time_format.

    convert_str : bool or dict, optional (default: True)
        If set to false, string values in the data columns will be kept.
        If set to true they will be converted to np.nan or 0 following default rules.
        If a dict is given mapping will be as given in the dict for values present in
        the dict and default as in parse_code for all other values

    copy_df : bool, optional (default: True)
        If set to true, a copy of the input DataFrame is made to keep the input as is.
        This negatively impacts speed. If set to false the input DataFrame will be
        altered but performance will be better

    Returns
    -------
    obj: pd.DataFrame
        pandas DataFrame with the read data

    Examples
    --------
    *Example for meta_mapping*::

        meta_mapping = {
            'pyCPA_col_1': {'col_1_value_1_in': 'col_1_value_1_out',
                            'col_1_value_2_in': 'col_1_value_2_out',
                            },
            'pyCPA_col_2': {'col_2_value_1_in': 'col_2_value_1_out',
                            'col_2_value_2_in': 'col_2_value_2_out',
                            },
        }

    *Example for filter_keep*::

        filter_keep = {
            'f_1': {'variable': ['CO2', 'CH4'], 'region': 'USA'},
            'f_2': {'variable': 'N2O'}
        }

    This example filter keeps all CO2 and CH4 data for the USA and N2O data for all
    countries

    *Example for filter_remove*::

        filter_remove = {
            'f_1': {'scenario': 'HISTORY'},
        }

    This filter removes all data with 'HISTORY' as scenario

    """
    # Check and prepare arguments
    if coords_defaults is None:
        coords_defaults = {}
    if add_coords_cols is None:
        add_coords_cols = {}
    if meta_data is None:
        attrs = {}
    else:
        attrs = meta_data.copy()

    check_mandatory_dimensions(coords_cols, coords_defaults)
    check_overlapping_specifications(coords_cols, coords_defaults)
    if add_coords_cols:
        check_overlapping_specifications_add_cols(coords_cols, add_coords_cols)

    # get all the columns that are actual data not metadata (usually the years)
    if time_cols is None:
        time_columns = [
            col
            for col in data_wide.columns.values
            if matches_time_format(col, time_format)
        ]
    else:
        time_columns = time_cols

    # if desired make a copy of the data to not alter the input data
    if copy_df:
        data_if = data_wide.copy(deep=True)
    else:
        data_if = data_wide

    filter_data(data_if, filter_keep, filter_remove)

    if convert_str:
        # get data columns (just one as we have long format)
        # find all string values
        str_values = find_str_values_in_data(data_if, time_columns)
        # create replacement dict
        str_repl_dict = create_str_replacement_dict(str_values, convert_str)
        replace_values(data_if, time_columns, str_repl_dict)

    add_dimensions_from_defaults(data_if, coords_defaults)

    naming_attrs = rename_columns(
        data_if, coords_cols, add_coords_cols, coords_defaults, coords_terminologies
    )
    attrs.update(naming_attrs)

    additional_coordinates = additional_coordinate_metadata(
        add_coords_cols, coords_cols, coords_terminologies
    )

    if coords_value_mapping is not None:
        map_metadata(data_if, attrs=attrs, meta_mapping=coords_value_mapping)

    if coords_value_filling is not None:
        data_if = fill_from_other_col(
            data_if, attrs=attrs, coords_value_filling=coords_value_filling
        )

    coords = list(set(data_if.columns.values) - set(time_columns))

    harmonize_units(data_if, dimensions=coords, attrs=attrs)

    data_if, coords = sort_columns_and_rows(data_if, dimensions=coords)
    dims = coords.copy()
    for add_coord in add_coords_cols.keys():
        dims.remove(add_coord)

    data_if.attrs = interchange_format_attrs_dict(
        xr_attrs=attrs,
        time_format=time_format,
        dimensions=dims,
        additional_coordinates=additional_coordinates,
    )

    return data_if


def read_wide_csv_file_if(
    filepath_or_buffer: str | Path | IO,
    *,
    coords_cols: dict[str, str],
    add_coords_cols: None | dict[str, list[str]] = None,
    coords_defaults: None | dict[str, Any] = None,
    coords_terminologies: dict[str, str],
    coords_value_mapping: None | dict[str, Any] = None,
    coords_value_filling: None | dict[str, dict[str, dict]] = None,
    filter_keep: None | dict[str, dict[str, Any]] = None,
    filter_remove: None | dict[str, dict[str, Any]] = None,
    meta_data: None | dict[str, Any] = None,
    time_format: str = "%Y",
    convert_str: bool | dict[str, float] = True,
) -> pd.DataFrame:
    """Read a CSV file in wide format into the PRIMAP2 interchange format.

    Columns can be renamed or filled with default values to match the PRIMAP2
    structure. Where we refer to "dimensions" in the parameter description below we
    mean the basic dimension names without the added terminology (e.g. "area" not
    "area (ISO3)"). The terminology information will be added by this function. You can
    not use the short dimension names in the attributes (e.g. "cat" instead of
    "category").

    TODO: Currently duplicate data points will not be detected.

    TODO: enable filtering through query strings

    TODO: enable specification of the entity terminology

    Parameters
    ----------
    filepath_or_buffer: str, pathlib.Path, or file-like
        Wide CSV file which will be read.

    coords_cols : dict
        Dict where the keys are PRIMAP2 dimensions and the values are column names in
        the files to be read. For secondary categories use a ``sec_cats__`` prefix.

    add_coords_cols : dict, optional
        Dict where the keys are PRIMAP2 additional coordinate names and the values are
        lists with two elements where the first is the column in the csv file to be
        read and the second is the primap2 dimension for the coordinate (e.g.
        ``category`` for a ``category_name`` coordinate.

    coords_defaults : dict, optional
        Dict for default values of coordinates / dimensions not given in the csv files.
        The keys are the dimension names and the values are the values for
        the dimensions. For secondary categories use a ``sec_cats__`` prefix.

    coords_terminologies : dict
        Dict defining the terminologies used for the different coordinates (e.g. ISO3
        for area). Only possible coordinates here are: area, category, scenario,
        entity, and secondary categories. For secondary categories use a ``sec_cats__``
        prefix. All entries different from "area", "category", "scenario", "entity",
        and ``sec_cats__<name>`` will raise a ValueError.

    coords_value_mapping : dict, optional
        A dict with primap2 dimension names as keys. Values are dicts with input values
        as keys and output values as values. A standard use case is to map gas names
        from input data to the standardized names used in primap2.
        Alternatively a value can also be a function which transforms one CSV metadata
        value into the new metadata value.
        A third possibility is to give a string as a value, which defines a rule for
        translating metadata values. The only defined rule at the moment is "PRIMAP1"
        which can be used for the "category", "entity", and "unit" columns to translate
        from PRIMAP1 metadata to PRIMAP2 metadata.

    coords_value_filling : dict, optional
        A dict with primap2 dimension names as keys. These are the target columns where
        values will be filled (or replaced). Vales are dicts with primap2 dimension
        names as keys. These are the source columns. The values are dicts with source
        value - target value mappings.
        The value filling can do everything that the value mapping can, but while
        mapping can only replace values within a column using information from that
        column, the filing function can also fill or replace data based on values from
        a different column.
        This can be used to e.g. fill missing category codes based on category names or
        to replace category codes which do not meet the terminology using the category
        names.

    filter_keep : dict, optional
        Dict defining filters of data to keep. Filtering is done before metadata
        mapping, so use original metadata values to define the filter. Column names are
        as in the csv file. Each entry in the dict defines an individual filter.
        The names of the filters have no relevance. Default: keep all data.

    filter_remove : dict, optional
        Dict defining filters of data to remove. Filtering is done before metadata
        mapping, so use original metadata values to define the filter. Column names are
        as in the csv file. Each entry in the dict defines an individual filter.
        The names of the filters have no relevance.

    meta_data : dict, optional
        Meta data for the whole dataset. Will end up in the dataset-wide attrs. Allowed
        keys are "references", "rights", "contact", "title", "comment", "institution",
        and "history". Documentation about the format and meaning of the meta data can
        be found in the
        `data format documentation <https://primap2.readthedocs.io/en/stable/\
data_format_details.html#dataset-attributes>`_.

    time_format : str, optional
        strftime style format used to parse the time information for the data columns.
        Default: "%Y", which will match years.

    convert_str : bool or dict, optional (default: True)
        If set to false, string values in the data columns will be kept.
        If set to true they will be converted to np.nan or 0 following default rules.
        If a dict is given mapping will be as given in the dict for values present in
        the dict and default as in parse_code for all other values

    Returns
    -------
    obj: pd.DataFrame
        pandas DataFrame with the read data

    Examples
    --------
    *Example for meta_mapping*::

        meta_mapping = {
            'pyCPA_col_1': {'col_1_value_1_in': 'col_1_value_1_out',
                            'col_1_value_2_in': 'col_1_value_2_out',
                            },
            'pyCPA_col_2': {'col_2_value_1_in': 'col_2_value_1_out',
                            'col_2_value_2_in': 'col_2_value_2_out',
                            },
        }

    *Example for filter_keep*::

        filter_keep = {
            'f_1': {'variable': ['CO2', 'CH4'], 'region': 'USA'},
            'f_2': {'variable': 'N2O'}
        }

    This example filter keeps all CO2 and CH4 data for the USA and N2O data for all
    countries

    *Example for filter_remove*::

        filter_remove = {
            'f_1': {'scenario': 'HISTORY'},
        }

    This filter removes all data with 'HISTORY' as scenario

    """
    # Check and prepare arguments
    if coords_defaults is None:
        coords_defaults = {}

    check_mandatory_dimensions(coords_cols, coords_defaults)
    check_overlapping_specifications(coords_cols, coords_defaults)
    if add_coords_cols:
        check_overlapping_specifications_add_cols(coords_cols, add_coords_cols)

    data, time_columns = read_wide_csv(
        filepath_or_buffer,
        coords_cols,
        add_coords_cols=add_coords_cols,
        time_format=time_format,
    )

    data = convert_wide_dataframe_if(
        data,
        coords_cols=coords_cols,
        add_coords_cols=add_coords_cols,
        coords_defaults=coords_defaults,
        coords_terminologies=coords_terminologies,
        coords_value_mapping=coords_value_mapping,
        coords_value_filling=coords_value_filling,
        filter_keep=filter_keep,
        filter_remove=filter_remove,
        meta_data=meta_data,
        time_format=time_format,
        time_cols=time_columns,
        convert_str=convert_str,
        copy_df=False,
    )

    return data


def interchange_format_attrs_dict(
    *,
    xr_attrs: dict,
    time_format: str,
    dimensions,
    additional_coordinates: None | dict = None,
) -> dict:
    metadata = {
        "attrs": xr_attrs,
        "time_format": time_format,
        "dimensions": {"*": dimensions.copy()},
    }

    if additional_coordinates:
        metadata["additional_coordinates"] = additional_coordinates

    return metadata


def additional_coordinate_metadata(
    add_coords_cols: dict[str, list[str]],
    coords_cols: dict[str, str],
    coords_terminologies: dict[str, str],
) -> dict:
    """Create the `additional_coordinates` dict and do a few consistency checks"""

    additional_coordinates = {}
    for coord in add_coords_cols:
        if coord in coords_terminologies:
            logger.error(
                f"Additional coordinate {coord} has terminology definition. "
                f"This is currently not supported by PRIMAP2."
            )
            raise ValueError(
                f"Additional coordinate {coord} has terminology definition. "
                f"This is currently not supported by PRIMAP2."
            )

        if add_coords_cols[coord][1] not in coords_cols:
            logger.error(
                f"Additional coordinate {coord} refers to unknown coordinate "
                f"{add_coords_cols[coord][1]}. "
            )
            raise ValueError(
                f"Additional coordinate {coord} refers to unknown coordinate "
                f"{add_coords_cols[coord][1]}. "
            )

        if add_coords_cols[coord][1] in coords_terminologies:
            additional_coordinates[coord] = (
                f"{add_coords_cols[coord][1]} "
                f"({coords_terminologies[add_coords_cols[coord][1]]})"
            )
        else:
            additional_coordinates[coord] = add_coords_cols[coord][1]

    return additional_coordinates


def check_mandatory_dimensions(
    coords_cols: dict[str, str],
    coords_defaults: dict[str, Any],
):
    """Check if all mandatory dimensions are specified."""
    for coord in INTERCHANGE_FORMAT_MANDATORY_COLUMNS:
        if coord not in coords_cols and coord not in coords_defaults:
            logger.error(
                f"Mandatory dimension {coord!r} not found in coords_cols={coords_cols}"
                f" or coords_defaults={coords_defaults}."
            )
            raise ValueError(f"Mandatory dimension {coord!r} not defined.")


def check_overlapping_specifications(
    coords_cols: dict[str, str],
    coords_defaults: dict[str, Any],
):
    both = set(coords_cols.keys()).intersection(set(coords_defaults.keys()))
    if both:
        logger.error(
            f"{both!r} is given in coords_cols and coords_defaults, but"
            f" it must only be given in one of them."
        )
        raise ValueError(f"{both!r} given in coords_cols and coords_defaults.")


def check_overlapping_specifications_add_cols(
    coords_cols: dict[str, str],
    add_coords_cols: dict[str, Any],
):
    cols_add = [val[0] for val in add_coords_cols.values()]
    both = set(coords_cols.values()).intersection(set(cols_add))
    if both:
        logger.error(
            f"columns {both!r} used for dimensions and additional coordinates, but"
            f" should be used in only one of them."
        )
        raise ValueError(f"{both!r} given in coords_cols and add_coords_cols.")


def matches_time_format(value: str, time_format: str) -> bool:
    try:
        datetime.datetime.strptime(value, time_format)
        return True
    except ValueError:
        return False


def read_wide_csv(
    filepath_or_buffer,
    coords_cols: dict[str, str],
    add_coords_cols: None | dict[str, list[str]] = None,
    time_format: str = "%Y",
) -> tuple[pd.DataFrame, list[str]]:
    data = pd.read_csv(
        filepath_or_buffer,
    )

    # get all the columns that are actual data not metadata (usually the years)
    time_cols = [
        col for col in data.columns.values if matches_time_format(col, time_format)
    ]

    # remove all cols not in the specification
    columns = data.columns.values
    if add_coords_cols:
        add_coords_col_names = {value[0] for value in add_coords_cols.values()}
    else:
        add_coords_col_names = set()

    data.drop(
        columns=list(
            set(columns)
            - set(coords_cols.values())
            - add_coords_col_names
            - set(time_cols)
        ),
        inplace=True,
    )

    # check that all cols in the specification could be read
    missing = set(coords_cols.values()) - set(data.columns.values)
    if missing:
        logger.error(
            f"Column(s) {missing} specified in coords_cols, but not found in "
            f"the CSV file {filepath_or_buffer!r}."
        )
        raise ValueError(f"Columns {missing} not found in CSV.")

    return data, time_cols


def read_long_csv(
    filepath_or_buffer,
    coords_cols: dict[str, str],
    add_coords_cols: None | dict[str, list[str]] = None,
) -> pd.DataFrame:
    if "data" not in coords_cols.keys():
        raise ValueError(
            "No data column in the CSV specified in coords_cols, so nothing to read."
        )

    if "time" in coords_cols:
        parse_dates = [coords_cols["time"]]
    else:
        parse_dates = False

    if add_coords_cols:
        add_coords_col_names = {value[0] for value in add_coords_cols.values()}
    else:
        add_coords_col_names = set()

    usecols = list(coords_cols.values()) + list(add_coords_col_names)

    data = pd.read_csv(
        filepath_or_buffer,
        parse_dates=parse_dates,
        usecols=usecols,
    )

    return data


def spec_to_query_string(filter_spec: dict[str, Any]) -> str:
    """Convert filter specification to query string.

    All column conditions in the filter are combined with &."""
    queries = []
    for col in filter_spec:
        if isinstance(filter_spec[col], list):
            itemlist = ", ".join(repr(x) for x in filter_spec[col])
            filter_query = f"`{col}` in [{itemlist}]"
        else:
            filter_query = f"`{col}` == {filter_spec[col]!r}"
        queries.append(filter_query)

    return " & ".join(queries)


def filter_data(
    data: pd.DataFrame,
    filter_keep: None | dict[str, dict[str, Any]] = None,
    filter_remove: None | dict[str, dict[str, Any]] = None,
):
    # Filters for keeping data are combined with "or" so that
    # everything matching at least one rule is kept.
    if filter_keep:
        queries = []
        for filter_spec in filter_keep.values():
            q = spec_to_query_string(filter_spec)
            queries.append(f"({q})")
        query = " | ".join(queries)
        data.query(query, inplace=True)

    # Filters for removing data are negated and combined with "and" so that
    # only rows which don't match any rule are kept.
    if filter_remove:
        queries = []
        for filter_spec in filter_remove.values():
            q = spec_to_query_string(filter_spec)
            queries.append(f"~({q})")
        query = " & ".join(queries)
        data.query(query, inplace=True)

    data.reset_index(drop=True, inplace=True)


def fill_from_other_col(
    df: pd.DataFrame,
    *,
    coords_value_filling: dict[str, dict[str, dict[str, str]]],
    attrs: dict[str, Any],
) -> pd.DataFrame:
    """
    This function fills value in one column based on values in other columns.
    It can be used to fill NaN values or to replace e.g. non-standard or
    non-unique category codes based on category names. It operates on pandas
    DataFrames.


    Parameters
    ----------
    df : pd.DataFrame
        Data to operate on

    coords_value_filling : dict
        A dict with primap2 dimension names as keys. These are the target columns where
        values will be filled (or replaced). Values are dicts with primap2 dimension
        names as keys. These are the source columns. The values are dicts with source
        value - target value mappings.
        This can be used to e.g. fill missing category codes based on category names or
        to replace category codes which do not meet the terminology using the category
        names.

    attrs : dict
        Dataset attributes

    Returns
    -------
    pd.DataFrame
    """
    dim_aliases = _selection.translations_from_attrs(attrs, include_entity=True)

    # loop over target columns in value mapping
    for target_col in coords_value_filling:
        target_info = coords_value_filling[target_col]
        # loop over source columns
        for source_col in target_info:
            mapping_info = target_info[source_col]
            # loop over cases
            target_col_name = dim_aliases.get(target_col, target_col)
            source_col_name = dim_aliases.get(source_col, source_col)
            for source_value in mapping_info:
                df.loc[df[source_col_name] == source_value, target_col_name] = df.loc[
                    df[source_col_name] == source_value, target_col_name
                ] = mapping_info[source_value]
    return df


def add_dimensions_from_defaults(
    data: pd.DataFrame,
    coords_defaults: dict[str, Any],
    additional_allowed_coords: Iterable[str] = (),
):
    if_columns = (
        INTERCHANGE_FORMAT_OPTIONAL_COLUMNS
        + INTERCHANGE_FORMAT_MANDATORY_COLUMNS
        + list(additional_allowed_coords)
    )
    for coord in coords_defaults.keys():
        if coord in if_columns or coord.startswith(SEC_CATS_PREFIX):
            # add column to dataframe with default value
            data[coord] = coords_defaults[coord]
        else:
            raise ValueError(
                f"{coord!r} given in coords_defaults is unknown - prefix with "
                f"{SEC_CATS_PREFIX!r} to add a secondary category."
            )


def map_metadata(
    data: pd.DataFrame,
    *,
    meta_mapping: dict[str, str | Callable | dict],
    attrs: dict[str, Any],
):
    """Map the metadata according to specifications given in meta_mapping.
    First map entity, then the rest."""
    meta_mapping_copy = meta_mapping.copy()
    if "entity" in meta_mapping.keys():
        meta_mapping_entity = dict(entity=meta_mapping_copy["entity"])
        meta_mapping_copy.pop("entity")
        map_metadata_unordered(data, meta_mapping=meta_mapping_entity, attrs=attrs)

    map_metadata_unordered(data, meta_mapping=meta_mapping_copy, attrs=attrs)


def map_metadata_unordered(
    data: pd.DataFrame,
    *,
    meta_mapping: dict[str, str | Callable | dict],
    attrs: dict[str, Any],
):
    """Map the metadata according to specifications given in meta_mapping."""
    dim_aliases = _selection.translations_from_attrs(attrs, include_entity=True)

    # TODO: add additional mapping functions here
    # values: (function, additional arguments)
    mapping_functions = {
        "PRIMAP1": {
            "category": (_conversion.convert_ipcc_code_primap_to_primap2, []),
            "entity": (_conversion.convert_entity_gwp_primap_to_primap2, []),
            "unit": (
                _conversion.convert_unit_to_primap2,
                [dim_aliases.get("entity", "entity")],
            ),
        }
    }

    meta_mapping_df = {}
    # preprocess meta_mapping
    for column, mapping in meta_mapping.items():
        column_name = dim_aliases.get(column, column)
        if isinstance(mapping, str) or callable(mapping):
            if isinstance(mapping, str):  # need to translate to function first
                try:
                    func, args = mapping_functions[mapping][column]
                except KeyError:
                    logger.error(
                        f"Unknown metadata mapping {mapping!r} for column {column!r}, "
                        f"known mappings are: {list(mapping_functions.keys())}."
                    )
                    raise ValueError(
                        f"Unknown metadata mapping {mapping!r} for column {column!r}."
                    ) from None
            else:
                func = mapping
                args = []

            if not args:  # simple case: no additional args needed
                values_to_map = data[column_name].unique()
                values_mapped = map(func, values_to_map)
                meta_mapping_df[column_name] = dict(
                    zip(values_to_map, values_mapped, strict=False)
                )

            else:  # need to supply additional arguments
                # this can't be handled using the replace()-call later since the
                # mapped values don't depend on the original values only, therefore
                # we do it directly
                sel = [column_name, *args]
                values_to_map = np.unique(data[sel].to_records(index=False))
                for vals_to_map in values_to_map:
                    # we replace values where all the arguments match - build a
                    # selector for that, then do the replacement
                    selector = data[column_name] == vals_to_map[0]
                    for i, arg in enumerate(args):
                        selector &= data[arg] == vals_to_map[i + 1]

                    data.loc[selector, column_name] = func(*vals_to_map)

        else:
            meta_mapping_df[column_name] = mapping

    data.replace(meta_mapping_df, inplace=True)


def rename_columns(
    data: pd.DataFrame,
    coords_cols: dict[str, str],
    add_coords_cols: dict[str, list[str]],
    coords_defaults: dict[str, Any],
    coords_terminologies: dict[str, str],
) -> dict:
    """Rename columns to match PRIMAP2 specifications and generate the corresponding
    dataset-wide attrs for PRIMAP2."""

    attr_names = {"category": "cat", "scenario": "scen", "area": "area"}

    attrs = {}
    sec_cats = []
    coord_renaming = {}

    for coord in itertools.chain(coords_cols, coords_defaults):
        if coord in coords_terminologies:
            name = f"{coord} ({coords_terminologies[coord]})"
            if coord == "entity":
                attrs["entity_terminology"] = coords_terminologies[coord]
        else:
            name = coord

        if coord.startswith(SEC_CATS_PREFIX):
            name = name[len(SEC_CATS_PREFIX) :]
            sec_cats.append(name)
        elif coord in attr_names:
            attrs[attr_names[coord]] = name

        coord_renaming[coords_cols.get(coord, coord)] = name

    for coord in add_coords_cols:
        coord_renaming[add_coords_cols[coord][0]] = coord

    data.rename(columns=coord_renaming, inplace=True)

    if sec_cats:
        attrs["sec_cats"] = sec_cats

    return attrs


_special_codes = {
    "C": np.nan,
    "nan": np.nan,
    "NaN": np.nan,
    "-": 0,
    "NE0": np.nan,
    "": np.nan,
}


def is_float(to_test: Any) -> bool:
    try:
        float(to_test)
        return True
    except ValueError:
        return False


def find_str_values_in_data(
    data: pd.DataFrame,
    columns: list[str],
) -> list:
    """Find all string values occurring in given columns of a DataFrame"""
    # limit our analysis to columns that contain strings
    # (or other object types)
    cols_with_strs = (
        data[columns].select_dtypes(include=[object]).columns.values.tolist()
    )
    temp = []
    for col in cols_with_strs:
        temp += list(data[col].unique())
    temp = list(set(temp))
    strs = [x for x in temp if not is_float(x)]
    return strs


def parse_code(code: str) -> float:
    """Parse a string code and return 0 or np.nan based on rules to interpret
    the codes."""
    code = code.strip()
    if code in _special_codes:
        return _special_codes[code]

    parts = code.split(",")
    parts = [x.replace(".", "").strip().upper() for x in parts]
    if "IE" in parts or "NO" in parts:
        return 0
    if "NE" in parts or "NA" in parts:
        return np.nan
    raise ValueError(f"Could not parse code: {code!r}.")


def create_str_replacement_dict(
    strs: list[str],
    user_str_conv: bool | dict[str, float],
) -> dict[str, str]:
    """Create a dict for replacement of strings by NaN and 0 based on
    general rules and user defined rules"""

    if isinstance(user_str_conv, bool):
        if user_str_conv:
            user_str_conv = {}
    elif isinstance(user_str_conv, dict):
        pass
    else:
        raise ValueError(
            f"Input for user_str_conv to create_str_replacement_dict"
            f" has to be a bool or a dict. {user_str_conv} is neither."
        )

    mapping = {}
    for str_val in strs:
        if str_val in user_str_conv:
            mapping[str_val] = user_str_conv[str_val]
        else:
            mapping[str_val] = parse_code(str_val)

    return mapping


def replace_values(data: pd.DataFrame, columns: list[str], na_repl_dict):
    """Replace str values indicating not-a-number by float NaN."""
    for col in columns:
        data[col] = data[col].replace(na_repl_dict)
        data[col] = pd.to_numeric(data[col], errors="coerce")
        data[col] = data[col].astype("float64", copy=False, errors="ignore")


def preferred_unit(entity: str, units: dict[str, str]) -> str | None:
    """Choose the preferred unit for the given entity.

    In general, "Gg <substance> / year" will be preferred if it is compatible with the
    given input units. Otherwise, the first unit from units will be preferred.

    Parameters
    ----------
    entity: str
        Basic entity, e.g. a gas.
    units: dict of str, str
        Keys are the units which are in use for the entity and values the global
        warming potential specifications which will be used for the conversion.
        By specifying the gwp_context to use, you can make sure that it will be possible to
        convert from the input units to the output unit with the given gwp.
        If not applicable for a unit use None

    Returns
    -------
    preferred_unit: str
        The best unit for the given entity and units.

    Examples
    --------
    >>> preferred_unit("CO2", {"kt CO2 / yr": None, "mg CO2 / s": None})
    'Gg CO2 / yr'
    >>> preferred_unit("CH4", {"kt CO2 / yr": "AR4GWP100"})
    'Gg CH4 / yr'
    >>> preferred_unit("CH4", {"kt CO2 / yr": None, "Mg CH4 / yr": None})
    None
    >>> preferred_unit("CH4", {"kt CO2 / yr": "AR4GWP100", "Gg CO2 / yr": "SARGWP100"})
    'Gg CH4 / yr'
    >>> preferred_unit(
    ...     "KYOTOGHG", {"kt CO2 / yr": "AR4GWP100", "Gg CO2 / yr": "SARGWP100"}
    ... )

    """
    unit_fallback = next(iter(units.keys()))
    context_fallback = units[unit_fallback]
    # check if all can be converted to native or fallback units
    native_conv = []
    fb_conv = []
    native_unit = "Gg " + entity + " / yr"
    for unit in units.keys():
        conversion_contexts = []
        if units[unit] is not None:
            conversion_contexts.append(units[unit])

        # check if conversion to native unit is possible
        try:
            # print(f"Testing conversion from {ureg[unit_fallback].units} to "
            #       f"{ureg[native_unit].units} for {entity}.")
            if ureg(unit).is_compatible_with(ureg[native_unit], *conversion_contexts):
                native_conv.append(True)
            else:
                native_conv.append(False)
        except pint.UndefinedUnitError:
            # we have a gas basket or something unknown, so no conversion to native unit
            # print(f"Exception occurred for entity {entity}")
            native_conv.append(False)
            pass

        # check if conversion to fallback unit is possible
        if units[unit] != context_fallback:
            fb_conv.append(False)
            pass
        else:
            try:
                # print(f"Testing conversion from {ureg[unit_fallback].units} to "
                #       f"{ureg[native_unit].units} for {entity}.")
                if ureg(unit).is_compatible_with(
                    ureg[unit_fallback], *conversion_contexts
                ):
                    fb_conv.append(True)
                else:
                    fb_conv.append(False)
            except pint.UndefinedUnitError:
                # we have a gas basket or something unknown, so no conversion to native
                # unit
                # print(f"Exception occurred for entity {entity}")
                fb_conv.append(False)
                pass

    if all(native_conv):
        # print(f"converting {entity} to native unit {native_unit}")
        return native_unit
    elif all(fb_conv):
        # print(f"converting {entity} to fallback unit {unit_fallback}")
        return unit_fallback
    else:
        return None


def harmonize_units(
    data: pd.DataFrame,
    *,
    unit_col: None | str = None,
    attrs: None | dict = None,
    dimensions: Iterable[str],
) -> None:
    """Harmonize the units of the input data.

    For each entity, convert all time series to the same unit (the native unit or
    the unit that occurs first if conversion to the native unit is not possible). Units
    must already be in PRIMAP2 style.

    As unit handling is tricky and with new units new problem occur this function has a
    lot of (currently commented) debug output

    Parameters
    ----------
    data: pd.DataFrame
        data for which the units should be harmonized
    unit_col: str, optional
        column name for unit column. Default: "unit"
    attrs: dict, optional
        attrs defining the aliasing of columns. If attrs contains "entity_terminology",
        "entity (<entity_terminology>)" will be used as the entity column, otherwise
        simply "entity" will be used as the entity column.
    dimensions: list of str
        the dimensions, i.e. the metadata columns.

    Returns
    -------
    None
        The data is altered in place.
    """
    # we need to convert the data such that we have one unit per entity
    data_cols = list(set(data.columns.values) - set(dimensions))

    if attrs is not None:
        dim_aliases = _selection.translations_from_attrs(attrs, include_entity=True)
        entity_col = dim_aliases.get("entity", "entity")
    else:
        entity_col = "entity"

    if unit_col is None:
        unit_col = dim_aliases.get("unit", "unit")

    # find basic entities for all entities and make a list
    entities = data[entity_col].unique()
    basic_entities = {}

    # print(entities)
    for entity in entities:
        # check if GWP given in entity
        # print(f"entity: {entity}")
        gwp_match = re.findall(r"\(([A-Z0-9]*)\)$", entity)
        if gwp_match:
            gwp_to_use = gwp_match[0]
            basic_entity = re.findall(r"^[^\(\)\s]*", entity)
            basic_entity = basic_entity[0]
        else:
            gwp_to_use = None
            basic_entity = entity
        # print(f"gwp: {gwp_to_use}")
        # print(f"basic_entity: {basic_entity}")
        if basic_entity in basic_entities.keys():
            basic_entities[basic_entity][entity] = gwp_to_use
        else:
            basic_entities[basic_entity] = {entity: gwp_to_use}

    for basic_entity in basic_entities:
        # print(f"basic_entity: {basic_entity}")
        # print(f"entities: {basic_entities[basic_entity]}")
        # get all units for this entity
        data_this_basic_entity = data.loc[
            data[entity_col].isin(basic_entities[basic_entity])
        ]
        units_this_basic_entity = data_this_basic_entity[unit_col].unique()
        unit_gwp_this_basic_entity = {}
        gwp_conversion_this_basic_entity = False
        for entity in basic_entities[basic_entity].keys():
            data_this_entity = data_this_basic_entity.loc[data[entity_col] == entity]
            units_this_entity = data_this_entity[unit_col].unique()
            for unit in units_this_entity:
                unit_gwp_this_basic_entity[unit] = basic_entities[basic_entity][entity]
            if basic_entities[basic_entity][entity] is not None:
                gwp_conversion_this_basic_entity = True

            if len(units_this_basic_entity) > 1 or gwp_conversion_this_basic_entity:
                # need unit conversion.
                # determine unit to convert all units to. If none is found no conversion
                # is carried out at all
                unit_to = preferred_unit(basic_entity, unit_gwp_this_basic_entity)
                # print(f"basic_entity: {basic_entity}, unit_to: {unit_to}")
                if unit_to is not None:
                    # print(f"unit conversion for {basic_entity}, "
                    #      f"{basic_entities[basic_entity]}")
                    for entity in basic_entities[basic_entity]:
                        data_this_entity = data.loc[data[entity_col] == entity]
                        units_this_entity = data_this_entity[unit_col].unique()

                        for unit in units_this_entity:
                            if unit != unit_to:
                                # print(f"Working on unit {unit}")
                                unit_pint = ureg[unit]
                                # could add a try except block here to throw and log an
                                # error or add error info in DF instead of crashing
                                gwp_this_entity = basic_entities[basic_entity][entity]
                                if gwp_this_entity:
                                    with ureg.context(gwp_this_entity):
                                        unit_pint = unit_pint.to(unit_to)
                                else:
                                    unit_pint = unit_pint.to(unit_to)
                                # print(f"Pint unit is {unit_pint}")
                                factor = unit_pint.magnitude
                                # print(f"Converting with factor {factor} to unit
                                # {unit_to}")
                                mask = (data[entity_col] == entity) & (
                                    data[unit_col] == unit
                                )
                                # print(data.loc[mask, data_cols])
                                try:
                                    data.loc[mask, data_cols] *= factor
                                except TypeError:
                                    # print(data.loc[mask, data_cols])
                                    strs = find_str_values_in_data(data, data_cols)
                                    logger.error(
                                        f"The following string values are present and "
                                        f"can not be converted during unit conversion: "
                                        f"{strs}."
                                    )
                                    raise ValueError(
                                        f"String values {strs} prevent unit conversion."
                                    ) from None

                                data.loc[mask, unit_col] = unit_to

                        # if entity differs from basic entity and the units are not
                        # compatible we had GWP conversion and have to adapt the entity
                        if (entity != basic_entity) and not ureg(
                            unit
                        ).is_compatible_with(ureg[unit_to]):
                            # entity was converted
                            entity_mask = data[entity_col] == entity
                            # print(f"Changing entity from {entity} to {basic_entity}")
                            data.loc[entity_mask, entity_col] = basic_entity


def sort_columns_and_rows(
    data: pd.DataFrame,
    dimensions: Iterable[str],
) -> tuple[pd.DataFrame, list[str]]:
    """Sort the data.

    The columns are ordered according to the order in
    INTERCHANGE_FORMAT_COLUMN_ORDER, with secondary categories alphabetically after
    the category and all date columns in order at the end.

    The rows are ordered by values of the non-date columns.

    Parameters
    ----------
    data: pd.DataFrame
        data which should be ordered
    dimensions: list of str
        the dimensions, i.e. the metadata columns.

    Returns
    -------
    sorted, dimensions_sorted : (pd.DataFrame, list of str)
        the input data frame with columns and rows ordered and the dimensions sorted.
    """
    time_cols = list(set(data.columns.values) - set(dimensions))

    other_cols = list(dimensions)
    cols_sorted = []
    for col in INTERCHANGE_FORMAT_COLUMN_ORDER:
        for ocol in other_cols:
            if ocol == col or (isinstance(ocol, str) and ocol.startswith(f"{col} (")):
                cols_sorted.append(ocol)
                other_cols.remove(ocol)
                break

    cols_sorted += list(sorted(other_cols))

    data: pd.DataFrame = data[cols_sorted + list(sorted(time_cols))]

    data.sort_values(by=cols_sorted, inplace=True)
    data.reset_index(inplace=True, drop=True)

    return data, cols_sorted



"""Data reading module of the PRIMAP2 climate policy analysis package."""

from ._data_reading import (
    convert_long_dataframe_if,
    convert_wide_dataframe_if,
    read_long_csv_file_if,
    read_wide_csv_file_if,
)
from ._GHG_inventory_reading import nir_add_unit_information, nir_convert_df_to_long
from ._interchange_format import (
    from_interchange_format,
    read_interchange_format,
    write_interchange_format,
)

__all__ = [
    "read_long_csv_file_if",
    "read_wide_csv_file_if",
    "convert_long_dataframe_if",
    "convert_wide_dataframe_if",
    "from_interchange_format",
    "read_interchange_format",
    "write_interchange_format",
    "nir_convert_df_to_long",
    "nir_add_unit_information",
]



import logging

import pytest
import xarray as xr
from loguru import logger

import primap2  # noqa: F401

from . import examples


# monkey-patch caplog to work wit loguru
# see https://loguru.readthedocs.io/en/stable/resources/migration.html#making-things-work-with-pytest-and-caplog
@pytest.fixture
def caplog(caplog):
    class PropogateHandler(logging.Handler):
        def emit(self, record):
            logging.getLogger(record.name).handle(record)

    handler_id = logger.add(PropogateHandler(), format="{message} {extra}")
    yield caplog
    logger.remove(handler_id)


@pytest.fixture
def minimal_ds() -> xr.Dataset:
    """A valid, minimal dataset."""
    return examples._cached_minimal_ds.copy(deep=True)


@pytest.fixture
def opulent_ds() -> xr.Dataset:
    """A valid dataset using lots of features."""
    return examples._cached_opulent_ds.copy(deep=True)


@pytest.fixture
def opulent_str_ds() -> xr.Dataset:
    """Like the opulent dataset, but additionally with a stringly typed data variable
    "method"."""
    return examples._cached_opulent_str_ds.copy(deep=True)


@pytest.fixture
def empty_ds() -> xr.Dataset:
    """An empty hull of a dataset with missing data."""
    return examples._cached_empty_ds.copy(deep=True)


@pytest.fixture
def opulent_processing_ds() -> xr.Dataset:
    """Like the opulent dataset, but additionally with processing information."""
    return examples._cached_opulent_processing_ds.copy(deep=True)


@pytest.fixture(
    params=["opulent", "opulent_str", "opulent_processing", "minimal", "empty"]
)
def any_ds(request) -> xr.Dataset:
    """Test with all available valid example Datasets."""
    if request.param == "opulent":
        return examples._cached_opulent_ds.copy(deep=True)
    elif request.param == "opulent_str":
        return examples._cached_opulent_str_ds.copy(deep=True)
    elif request.param == "opulent_processing":
        return examples._cached_opulent_processing_ds.copy(deep=True)
    elif request.param == "minimal":
        return examples._cached_minimal_ds.copy(deep=True)
    elif request.param == "empty":
        return examples._cached_empty_ds.copy(deep=True)



"""Unit handling based on pint, pint_xarray and openscm_units.

Portions of this file are copied from pint_xarray and are
Copyright 2020, pint-xarray developers."""

import pint
import pint_xarray
import xarray as xr
from openscm_units import unit_registry as ureg

from . import _accessor_base

pint_xarray.setup_registry(ureg)


# HACK to fix https://github.com/hgrecco/pint/issues/1957
def _fix_super_slow_pint_ndim():
    import numbers

    import pint.facets

    def faster_ndim(self) -> int:
        if isinstance(self.magnitude, numbers.Number):
            return 0
        if str(type(self.magnitude)) == "NAType":
            return 0
        return self.magnitude.ndim

    setattr(pint.facets.plain.quantity.PlainQuantity, "ndim", property(faster_ndim))  # noqa


if pint.__version__ == "0.23":
    _fix_super_slow_pint_ndim()


class DataArrayUnitAccessor(_accessor_base.BaseDataArrayAccessor):
    """Provide functions for unit handling"""

    def quantify(self, **kwargs):
        """Attaches units to the DataArray.

        Units can be specified as a pint.Unit or as a string. If no units are specified
        then the
        units will be parsed from the `'units'` entry of the DataArray's
        `.attrs`. Will raise a ValueError if the DataArray already contains a
        unit-aware array.

        This function is a wrapper for pint_xarrays function with the
        same name, which uses the primap2 unit registry.
        Calling ``da.pr.quantify()`` is therefore equivalent to calling
        ``da.pint.quantify(unit_registry=primap2.ureg)``

        .. note::
            Be aware that unless you're using ``dask`` this will load
            the data into memory. To avoid that, consider converting
            to ``dask`` first (e.g. using ``chunk``).

            As units in dimension coordinates are not supported until
            ``xarray`` changes the way it implements indexes, these
            units will be set as attributes.

        Parameters
        ----------
        units : unit-like or mapping of hashable to unit-like, optional
            Physical units to use for this DataArray. If a str or
            pint.Unit, will be used as the DataArray's units. If a
            dict-like, it should map a variable name to the desired
            unit (use the DataArray's name to refer to its data). If
            not provided, will try to read them from
            ``DataArray.attrs['units']`` using pint's parser. The
            ``"units"`` attribute will be removed from all variables
            except from dimension coordinates.
        **unit_kwargs
            Keyword argument form of units.

        Returns
        -------
        quantified : DataArray
            DataArray whose wrapped array data will now be a Quantity
            array with the specified units.

        Examples
        --------
        >>> da = xr.DataArray(
        ...     data=[0.4, 0.9, 1.7, 4.8, 3.2, 9.1],
        ...     dims=["wavelength"],
        ...     coords={"wavelength": [1e-4, 2e-4, 4e-4, 6e-4, 1e-3, 2e-3]},
        ... )
        >>> da.pint.quantify(units="Hz")
        <xarray.DataArray (wavelength: 6)> Size: 48B
        <Quantity([0.4 0.9 1.7 4.8 3.2 9.1], 'hertz')>
        Coordinates:
          * wavelength  (wavelength) float64 48B 0.0001 0.0002 0.0004 0.0006 0.001 0.002
        """
        return self._da.pint.quantify(unit_registry=ureg, **kwargs)

    def dequantify(self) -> xr.DataArray:
        """Removes units from the DataArray and its coordinates.

        Will replace ``.attrs['units']`` on each variable with a string
        representation of the ``pint.Unit`` instance.

        Returns
        -------
        dequantified : DataArray
            DataArray whose array data is unitless, and of the type
            that was previously wrapped by `pint.Quantity`.
        """
        return self._da.pint.dequantify()

    def convert_to_gwp(self, gwp_context: str, units: str | pint.Unit) -> xr.DataArray:
        """Convert to a global warming potential

        Parameters
        ----------
        gwp_context: str
            The global warming potential context to use for the conversion, as
            understood by ``openscm_units``.
        units: str or pint unit
            The units in which the global warming potential is given after the
            conversion.

        Returns
        -------
        converted : xr.DataArray
        """
        if (
            "gwp_context" in self._da.attrs
            and self._da.attrs["gwp_context"] != gwp_context
        ):
            raise ValueError(
                f"Incompatible gwp conversions: {self._da.attrs['gwp_context']!r}"
                f" != {gwp_context!r}."
            )

        with ureg.context(gwp_context):
            da = self._da.pint.to(units)
        da.attrs["gwp_context"] = gwp_context
        da.name = f"{da.attrs['entity']} ({da.attrs['gwp_context']})"
        return da

    def convert_to_gwp_like(self, like: xr.DataArray) -> xr.DataArray:
        """Convert to a global warming potential in the units of a reference array
        using the ``gwp_context`` of the reference array.

        Parameters
        ----------
        like: xr.DataArray
            Other DataArray containing a global warming potential.

        Returns
        -------
        converted : xr.DataArray
        """
        if "gwp_context" not in like.attrs or like.attrs["gwp_context"] is None:
            raise ValueError("reference array has no gwp_context.")
        if like.pint.units is None:
            raise ValueError("reference array has no units attached.")
        return self.convert_to_gwp(
            gwp_context=like.attrs["gwp_context"], units=like.pint.units
        )

    @property
    def gwp_context(self) -> pint.Context:
        """The pint conversion context for this DataArray, directly usable for
        conversions.

        Examples
        --------
        >>> import primap2
        >>> import primap2.tests
        >>> ds = primap2.tests.minimal_ds()
        >>> with ds["SF6 (SARGWP100)"].pr.gwp_context:
        ...     ds["CH4"].pint.to("Gg CO2 / year")

        Returns
        -------
        context : pint.Context
        """
        return ureg.context(self._da.attrs["gwp_context"])

    def convert_to_mass(
        self, gwp_context: str | None = None, entity: str | None = None
    ) -> xr.DataArray:
        """Convert a global warming potential of a greenhouse gas to a mass.

        Parameters
        ----------
        gwp_context: str, optional
            The global warming potential context to be used for the conversion.
            It must be one of the global warming potential contexts understood by
            ``openscm_units``. If omitted, the global warming potential context used to
            calculate the global warming potential originally is used, so you should
            only need to provide an explicit gwp_context in exceptional cases.
        entity: str, optional
            The entity into which the global warming potential should be converted.
            If omitted, the original entity is used, so you should only need to provide
            an explicit entity in exceptional cases.

        Returns
        -------
        converted : xr.DataArray
        """
        if gwp_context is None:
            try:
                gwp_context = self._da.attrs["gwp_context"]
            except KeyError:
                raise ValueError(
                    "No gwp_context given and no gwp_context available in the attrs."
                ) from None
        if entity is None:
            try:
                entity = self._da.attrs["entity"]
            except KeyError:
                raise ValueError(
                    "No entity given and no entity available in the attrs."
                ) from None

        if isinstance(entity, str):
            entity = ureg.parse_units(entity)

        with ureg.context(gwp_context):
            da = self._da.pint.to(
                self._da.pint.units / ureg.parse_units("CO2") * entity
            )

        if "gwp_context" in da.attrs:
            del da.attrs["gwp_context"]
        da.attrs["entity"] = str(entity)
        da.name = str(entity)
        return da


class DatasetUnitAccessor(_accessor_base.BaseDatasetAccessor):
    """Provides functions for unit handling."""

    def quantify(self, units=None, **unit_kwargs) -> xr.Dataset:
        """Attaches units to each variable in the Dataset.

        Units can be specified as a ``pint.Unit`` or as a
        string. If no
        units are specified then the units will be parsed from the
        ``"units"`` entry of the Dataset variable's ``.attrs``. Will
        raise a ValueError if any of the variables already contain a
        unit-aware array.

        This function is a wrapper for pint_xarrays function with the
        same name, which uses the primap2 unit registry.
        Calling ``ds.pr.quantify()`` is therefore equivalent to calling
        ``ds.pint.quantify(unit_registry=primap2.ureg)``

        .. note::
            Be aware that unless you're using ``dask`` this will load
            the data into memory. To avoid that, consider converting
            to ``dask`` first (e.g. using ``chunk``).

            As units in dimension coordinates are not supported until
            ``xarray`` changes the way it implements indexes, these
            units will be set as attributes.

        Parameters
        ----------
        units : mapping of hashable to unit-like, optional
            Physical units to use for particular DataArrays in this
            Dataset. It should map variable names to units (unit names
            or ``pint.Unit`` objects). If not provided, will try to
            read them from ``Dataset[var].attrs['units']`` using
            pint's parser. The ``"units"`` attribute will be removed
            from all variables except from dimension coordinates.
        **unit_kwargs
            Keyword argument form of ``units``.

        Returns
        -------
        quantified : Dataset
            The variables in quantified will now contain Quantity arrays
            with units.

        Examples
        --------
        >>> import xarray as xr
        >>> import primap2
        >>> ds = xr.Dataset(
        ...     {"a": ("x", [0, 3, 2], {"units": "m"}), "b": ("x", [5, -2, 1])},
        ...     coords={"x": [0, 1, 2], "u": ("x", [-1, 0, 1], {"units": "s"})},
        ... )
        >>> ds
        <xarray.Dataset> Size: ...
        Dimensions:  (x: 3)
        Coordinates:
          * x        (x) int... 0 1 2
            u        (x) int... -1 0 1
        Data variables:
            a        (x) int... 0 3 2
            b        (x) int... 5 -2 1

        >>> ds.pr.quantify()
        <xarray.Dataset> Size: ...
        Dimensions:  (x: 3)
        Coordinates:
          * x        (x) int... 0 1 2
            u        (x) int... [s] -1 0 1
        Data variables:
            a        (x) int... [m] 0 3 2
            b        (x) int... 5 -2 1
        >>> ds.pr.quantify({"b": "dm"})
        <xarray.Dataset> Size: ...
        Dimensions:  (x: 3)
        Coordinates:
          * x        (x) int... 0 1 2
            u        (x) int... [s] -1 0 1
        Data variables:
            a        (x) int... [m] 0 3 2
            b        (x) int... [dm] 5 -2 1
        """
        return self._ds.pint.quantify(unit_registry=ureg, units=units, **unit_kwargs)

    def dequantify(self) -> xr.Dataset:
        """Removes units from the Dataset and its coordinates.

        Will replace ``.attrs['units']`` on each variable with a string
        representation of the ``pint.Unit`` instance.

        Returns
        -------
        dequantified : Dataset
            Dataset whose data variables are unitless, and of the type
            that was previously wrapped by ``pint.Quantity``.
        """
        return self._ds.pint.dequantify()



import datetime

import numpy as np
import pandas as pd
import xarray as xr

import primap2
from primap2 import ureg


def minimal_ds() -> xr.Dataset:
    """A valid, minimal dataset."""
    time = pd.date_range("2000-01-01", "2020-01-01", freq="YS")
    area_iso3 = np.array(["COL", "ARG", "MEX", "BOL"])

    # seed the rng with a constant to achieve predictable "randomness"
    rng = np.random.default_rng(1)

    minimal = xr.Dataset(
        {
            ent: xr.DataArray(
                data=rng.random((len(time), len(area_iso3), 1)),
                coords={
                    "time": time,
                    "area (ISO3)": area_iso3,
                    "source": ["RAND2020"],
                },
                dims=["time", "area (ISO3)", "source"],
                attrs={"units": f"{ent} Gg / year", "entity": ent},
            )
            for ent in ("CO2", "SF6", "CH4")
        },
        attrs={"area": "area (ISO3)"},
    ).pr.quantify()

    with ureg.context("SARGWP100"):
        minimal["SF6 (SARGWP100)"] = minimal["SF6"].pint.to("CO2 Gg / year")
    minimal["SF6 (SARGWP100)"].attrs["gwp_context"] = "SARGWP100"
    return minimal


COORDS = {
    "time": pd.date_range("2000-01-01", "2020-01-01", freq="YS"),
    "area (ISO3)": np.array(["COL", "ARG", "MEX", "BOL"]),
    "category (IPCC 2006)": np.array(["0", "1", "2", "3", "4", "5", "1.A", "1.B"]),
    "animal (FAOSTAT)": np.array(["cow", "swine", "goat"]),
    "product (FAOSTAT)": np.array(["milk", "meat"]),
    "scenario (FAOSTAT)": np.array(["highpop", "lowpop"]),
    "provenance": np.array(["projected"]),
    "model": np.array(["FANCYFAO"]),
    "source": np.array(["RAND2020", "RAND2021"]),
}


def opulent_ds() -> xr.Dataset:
    """A valid dataset using lots of features."""

    # seed the rng with a constant to achieve predictable "randomness"
    rng = np.random.default_rng(1)

    opulent = xr.Dataset(
        {
            ent: xr.DataArray(
                data=rng.random(tuple(len(x) for x in COORDS.values())),
                coords=COORDS,
                dims=list(COORDS.keys()),
                attrs={"units": f"{ent} Gg / year", "entity": ent},
            )
            for ent in ("CO2", "SF6", "CH4")
        },
        attrs={
            "entity_terminology": "primap2",
            "area": "area (ISO3)",
            "cat": "category (IPCC 2006)",
            "sec_cats": ["animal (FAOSTAT)", "product (FAOSTAT)"],
            "scen": "scenario (FAOSTAT)",
            "references": "doi:10.1012",
            "rights": "Use however you want.",
            "contact": "lol_no_one_will_answer@example.com",
            "title": "Completely invented GHG inventory data",
            "comment": "GHG inventory data ...",
            "institution": "PIK",
            "publication_date": datetime.date(2099, 12, 31),
        },
    )

    pop_coords = {
        x: COORDS[x]
        for x in (
            "time",
            "area (ISO3)",
            "provenance",
            "model",
            "source",
        )
    }
    pop_shape = tuple(len(x) for x in pop_coords.values())
    opulent["population"] = xr.DataArray(
        data=rng.random(pop_shape),
        coords=pop_coords,
        dims=list(pop_coords.keys()),
        attrs={"entity": "population", "units": ""},
    )

    opulent = opulent.assign_coords(
        {
            "category_names": xr.DataArray(
                data=np.array(
                    [
                        "total",
                        "industry",
                        "energy",
                        "transportation",
                        "residential",
                        "land use",
                        "heavy industry",
                        "light industry",
                    ]
                ),
                coords={"category (IPCC 2006)": COORDS["category (IPCC 2006)"]},
                dims=["category (IPCC 2006)"],
            )
        }
    )

    opulent = opulent.pint.quantify(unit_registry=ureg)

    with ureg.context("SARGWP100"):
        opulent["SF6 (SARGWP100)"] = opulent["SF6"].pint.to("CO2 Gg / year")
    opulent["SF6 (SARGWP100)"].attrs["gwp_context"] = "SARGWP100"

    return opulent


def opulent_str_ds() -> xr.Dataset:
    """Like the opulent dataset, but additionally with a stringly typed data variable
    "method"."""
    opulent = opulent_ds()

    method_coords = {
        x: COORDS[x]
        for x in (
            "time",
            "area (ISO3)",
            "model",
            "source",
        )
    }
    method_shape = tuple(len(x) for x in method_coords.values())
    opulent["method"] = xr.DataArray(
        data=np.ones(method_shape, dtype=str).astype(object),
        coords=method_coords,
        dims=list(method_coords.keys()),
        attrs={"entity": "method"},
    )
    opulent["method"].pr.loc[{"time": "2000", "area": "COL", "source": "RAND2020"}] = (
        "text"
    )

    return opulent


def opulent_processing_ds() -> xr.Dataset:
    """Like the opulent dataset, but additionally with processing information data
    variables."""

    opulent = opulent_ds()

    new_vars = {}
    for var in opulent.keys():
        dims = [dim for dim in opulent.dims.keys() if dim != "time"]
        shape = tuple(len(opulent[x]) for x in dims)
        new_vars[f"Processing of {var}"] = xr.DataArray(
            data=np.full(
                shape=shape,
                fill_value=primap2.TimeseriesProcessingDescription(
                    steps=[
                        primap2.ProcessingStepDescription(
                            time="all",
                            function="random",
                            description="Values created randomly.",
                        )
                    ]
                ),
            ),
            coords=opulent[dims],
            dims=dims,
            attrs={
                "entity": f"Processing of {var}",
                "described_variable": var,
            },
        )

    opulent.update(new_vars)

    return opulent


def empty_ds() -> xr.Dataset:
    """An empty hull of a dataset with missing data."""
    time = pd.date_range("2000-01-01", "2020-01-01", freq="YS")
    area_iso3 = np.array(["COL", "ARG", "MEX", "BOL"])
    coords = {
        "time": time,
        "area (ISO3)": area_iso3,
        "source": ["RAND2020"],
    }
    dims = ["time", "area (ISO3)", "source"]
    empty = xr.Dataset(
        {
            ent: xr.DataArray(
                data=np.zeros((len(time), len(area_iso3), 1), dtype=float),
                coords=coords,
                dims=dims,
                attrs={"units": f"{ent} Gg / year", "entity": ent},
            )
            for ent in ("CO2", "SF6", "CH4")
        },
        attrs={"area": "area (ISO3)"},
    ).pr.quantify()

    empty["KYOTOGHG (AR4GWP100)"] = xr.DataArray(
        data=np.zeros((len(time), len(area_iso3), 1), dtype=float),
        coords=coords,
        dims=dims,
        attrs={
            "units": "CO2 Gg / year",
            "entity": "KYOTOGHG",
            "gwp_context": "AR4GWP100",
        },
    ).pr.quantify()

    return empty


_cached_minimal_ds = minimal_ds()
_cached_opulent_ds = opulent_ds()
_cached_opulent_str_ds = opulent_str_ds()
_cached_opulent_processing_ds = opulent_processing_ds()
_cached_empty_ds = empty_ds()



"""Composite Source Generator

Generate a composite harmonized dataset from multiple sources according to defined
source priorities and matching algorithms.
"""

from ._compose import compose
from ._models import (
    PriorityDefinition,
    StrategyDefinition,
    StrategyUnableToProcess,
)
from ._strategies.global_least_squares import GlobalLSStrategy
from ._strategies.substitution import SubstitutionStrategy

__all__ = [
    "compose",
    "PriorityDefinition",
    "StrategyDefinition",
    "SubstitutionStrategy",
    "StrategyUnableToProcess",
    "GlobalLSStrategy",
]



#!/usr/bin/env python
"""Tests for _merge.py"""

import pytest
import xarray as xr

from .utils import assert_aligned_equal, assert_ds_aligned_equal


def test_merge_disjoint_vars(opulent_ds):
    ds_start: xr.Dataset = opulent_ds[["CO2"]]
    ds_merge: xr.Dataset = opulent_ds[["CH4"]]
    ds_result = ds_start.pr.merge(ds_merge)

    assert_ds_aligned_equal(ds_result, opulent_ds[["CH4", "CO2"]])


def test_merge_ds_pass_tolerance(opulent_ds):
    sel = {"time": slice("2000", "2005"), "scenario": "highpop", "source": "RAND2020"}
    ds = opulent_ds.pr.loc[sel]
    ds_start = ds.pr.loc[{"area": ["ARG", "COL"]}]
    ds_merge = ds.pr.loc[{"area": ["ARG", "MEX"]}]

    data_to_modify = ds["CO2"].pr.loc[{"area": ["ARG"]}].pr.sum("area")
    data_to_modify.data *= 1.009

    da_merge = ds["CO2"].pr.set("area", "ARG", data_to_modify, existing="overwrite")
    ds_merge["CO2"] = da_merge
    ds_result = ds_start.pr.merge(ds_merge, tolerance=0.01)

    assert_ds_aligned_equal(ds_result, ds.pr.loc[{"area": ["ARG", "COL", "MEX"]}])


def test_merge_pass_tolerance(opulent_ds):
    # only take part of the countries to have something to actually merge
    da_start = opulent_ds["CO2"].pr.loc[{"area": ["ARG", "COL", "MEX"]}]
    data_to_modify = opulent_ds["CO2"].pr.loc[{"area": ["ARG"]}].pr.sum("area")
    data_to_modify.data *= 1.009
    da_merge = opulent_ds["CO2"].pr.set(
        "area", "ARG", data_to_modify, existing="overwrite"
    )
    da_result = da_start.pr.merge(da_merge, tolerance=0.01)

    assert_aligned_equal(da_result, opulent_ds["CO2"])


def test_merge_fail_tolerance(opulent_ds):
    da_start = opulent_ds["CO2"]
    data_to_modify = opulent_ds["CO2"].pr.loc[{"area": ["ARG"]}].pr.sum("area")
    data_to_modify.data = data_to_modify.data * 1.09
    da_merge = opulent_ds["CO2"].pr.set(
        "area", "ARG", data_to_modify, existing="overwrite"
    )

    with pytest.raises(
        xr.MergeError,
        match="pr.merge error: found discrepancies " "larger than tolerance",
    ):
        da_start.pr.merge(da_merge, tolerance=0.01)


def test_merge_fail_tolerance_warn(opulent_ds):
    da_start = opulent_ds["CO2"]
    data_to_modify = opulent_ds["CO2"].pr.loc[{"area": ["ARG"]}].pr.sum("area")
    data_to_modify.data = data_to_modify.data * 1.09
    da_merge = opulent_ds["CO2"].pr.set(
        "area", "ARG", data_to_modify, existing="overwrite"
    )

    da_result = da_start.pr.merge(da_merge, tolerance=0.01, error_on_discrepancy=False)
    assert_aligned_equal(da_result, da_start)


def test_coords_not_matching_ds(opulent_ds):
    ds_start = opulent_ds
    ds_merge = opulent_ds.rename({"time": "year"})

    with pytest.raises(
        ValueError, match="pr.merge error: coords of objects to merge must agree"
    ):
        ds_start.pr.merge(ds_merge)


def test_coords_not_matching_da(opulent_ds):
    da_start = opulent_ds["CH4"]
    da_merge = opulent_ds["CH4"].rename({"time": "year"})

    with pytest.raises(
        ValueError, match="pr.merge error: coords of objects to merge must agree"
    ):
        da_start.pr.merge(da_merge)


def test_dims_not_matching_ds(opulent_ds):
    ds_start = opulent_ds
    ds_merge = opulent_ds.rename_dims({"time": "year"})

    with pytest.raises(
        ValueError, match="pr.merge error: dims of objects to merge must agree"
    ):
        ds_start.pr.merge(ds_merge)


def test_dims_not_matching_da(opulent_ds):
    da_start = opulent_ds["CH4"]
    da_merge = opulent_ds["CH4"].swap_dims({"time": "year"})

    with pytest.raises(
        ValueError, match="pr.merge error: dims of objects to merge must agree"
    ):
        da_start.pr.merge(da_merge)


# tests to check if additional coords stay in place during merge. Different setups as pr.merge
# uses different methods depending on the setup. as xr.merge is not affected by the coordinate
# bug these tests could also be removed. But if kept they detect if the bug appears for xr.merge
def test_merge_ds_add_coord(opulent_ds):
    country_names = ["Colombia", "Argentina", "Mexico", "Bolivia"]
    ds_full = opulent_ds.assign_coords(country_name=("area (ISO3)", country_names))
    sel = {"time": slice("2000", "2005"), "scenario": "highpop", "source": "RAND2020"}
    ds = ds_full.pr.loc[sel]
    ds_start = ds.pr.loc[{"area": ["ARG", "COL"]}]
    ds_merge = ds.pr.loc[{"area": ["ARG", "MEX"]}]

    ds_result = ds_start.pr.merge(ds_merge, tolerance=0.01)
    ds_compare = ds.pr.loc[{"area (ISO3)": ["ARG", "COL", "MEX"]}]

    # check that the additional coord is present in result
    assert "country_name" in list(ds_result.coords)
    # check that the mapping of country names to country codes is intact
    # (meaning that the additional coordinate is aligned correctly)
    for country in ds_compare.coords["area (ISO3)"].values:
        assert (
            ds_result.coords["country_name"].loc[{"area (ISO3)": country}]
            == ds_compare.coords["country_name"].loc[{"area (ISO3)": country}]
        )


def test_merge_ds_add_coord_tolerance(opulent_ds):
    country_names = ["Colombia", "Argentina", "Mexico", "Bolivia"]
    ds_full = opulent_ds.assign_coords(country_name=("area (ISO3)", country_names))
    sel = {"time": slice("2000", "2005"), "scenario": "highpop", "source": "RAND2020"}
    ds = ds_full.pr.loc[sel]
    ds_start = ds.pr.loc[{"area": ["ARG", "COL"]}]
    ds_merge = ds.pr.loc[{"area": ["ARG", "MEX"]}]

    data_to_modify = ds["CO2"].pr.loc[{"area": ["ARG"]}].pr.sum("area")
    data_to_modify.data *= 1.009

    da_merge = ds["CO2"].pr.set("area", "ARG", data_to_modify, existing="overwrite")
    ds_merge["CO2"] = da_merge
    ds_result = ds_start.pr.merge(ds_merge, tolerance=0.01)
    ds_compare = ds.pr.loc[{"area (ISO3)": ["ARG", "COL", "MEX"]}]

    # check that the additional coord is present in result
    assert "country_name" in list(ds_result.coords)
    # check that the mapping of country names to country codes is intact
    # (meaning that the additional coordinate is aligned correctly)
    for country in ds_compare.coords["area (ISO3)"].values:
        assert (
            ds_result.coords["country_name"].loc[{"area (ISO3)": country}]
            == ds_compare.coords["country_name"].loc[{"area (ISO3)": country}]
        )


def test_merge_ds_add_coord_disjoint_vars(opulent_ds):
    country_names = ["Colombia", "Argentina", "Mexico", "Bolivia"]
    ds_full = opulent_ds.assign_coords(country_name=("area (ISO3)", country_names))

    ds_start: xr.Dataset = ds_full[["CO2"]]
    ds_merge: xr.Dataset = ds_full[["CH4"]]
    ds_result = ds_start.pr.merge(ds_merge)
    ds_compare = ds_full[["CO2", "CH4"]]

    # check that the additional coord is present in result
    assert "country_name" in list(ds_merge.coords)
    # check that the mapping of country names to country codes is intact
    # (meaning that the additional coordinate is aligned correctly)
    for country in ds_compare.coords["area (ISO3)"].values:
        assert (
            ds_result.coords["country_name"].loc[{"area (ISO3)": country}]
            == ds_compare.coords["country_name"].loc[{"area (ISO3)": country}]
        )



import datetime

from . import _accessor_base


class DatasetMetadataAccessor(_accessor_base.BaseDatasetAccessor):
    @property
    def entity_terminology(self) -> str:
        """the terminology of the entity attributes in this dataset's data variables"""
        return self._ds.attrs["entity_terminology"]

    @entity_terminology.setter
    def entity_terminology(self, value: str):
        self._ds.attrs["entity_terminology"] = value

    @property
    def references(self) -> str:
        """citable reference(s) describing the data

        If the references start with ``doi:``, it is a doi, otherwise it is a
        free-form literature reference.
        """
        return self._ds.attrs["references"]

    @references.setter
    def references(self, value: str):
        self._ds.attrs["references"] = value

    @property
    def rights(self) -> str:
        """license or other usage restrictions of the data"""
        return self._ds.attrs["rights"]

    @rights.setter
    def rights(self, value: str):
        self._ds.attrs["rights"] = value

    @property
    def contact(self) -> str:
        """who can answer questions about the data"""
        return self._ds.attrs["contact"]

    @contact.setter
    def contact(self, value: str):
        self._ds.attrs["contact"] = value

    @property
    def title(self) -> str:
        """a succinct description"""
        return self._ds.attrs["title"]

    @title.setter
    def title(self, value: str):
        self._ds.attrs["title"] = value

    @property
    def comment(self) -> str:
        """longer form description"""
        return self._ds.attrs["comment"]

    @comment.setter
    def comment(self, value: str):
        self._ds.attrs["comment"] = value

    @property
    def institution(self) -> str:
        """where the data originates"""
        return self._ds.attrs["institution"]

    @institution.setter
    def institution(self, value: str):
        self._ds.attrs["institution"] = value

    @property
    def publication_date(self) -> datetime.date:
        """The original date of publication of the dataset, if published."""
        return self._ds.attrs["publication_date"]

    @publication_date.setter
    def publication_date(self, value: datetime.date):
        self._ds.attrs["publication_date"] = value



"""The PRIMAP2 climate policy analysis package."""

__author__ = """Mika Pflüger"""
__email__ = "mika.pflueger@pik-potsdam.de"
__version__ = "0.11.1"

from . import accessors, pm2io
from ._data_format import (
    ProcessingStepDescription,
    TimeseriesProcessingDescription,
    open_dataset,
)
from ._selection import Not
from ._units import ureg

__all__ = [
    "accessors",
    "open_dataset",
    "ureg",
    "pm2io",
    "ProcessingStepDescription",
    "TimeseriesProcessingDescription",
    "Not",
]



"""Compose a harmonized dataset from multiple input datasets."""

import contextlib
import math
import typing
from collections.abc import Hashable

import numpy as np
import tqdm
import xarray as xr
from loguru import logger

import primap2._data_format

from . import _models


def compose(
    *,
    input_data: xr.Dataset,
    priority_definition: _models.PriorityDefinition,
    strategy_definition: _models.StrategyDefinition,
    progress_bar: type[tqdm.tqdm] | None = tqdm.tqdm,
) -> xr.Dataset:
    """
    Compose a harmonized dataset from multiple input datasets.

    The input datasets are treated at the timeseries level, for each timeseries:

    * the highest priority dataset is chosen according to the priority_definition
    * if the dataset contains any nan values (denoting missing information), the next
      highest priority dataset is selected to fill the missing values
    * a filling strategy is chosen according to the strategy_definition and the filling
      is done with the dataset selected in the previous step
    * if afterwards there are still missing nan values, the next highest priority
      dataset is selected and the previous step is repeated
    * when all missing values are filled or all input datasets are exhausted, the
      timeseries is finished and written to the result.

    In addition to the harmonized data, also a description of the processing steps
    done for each timeseries is returned in the result dataset, where for each
    variable, a variable of the form "Processing of $variable" is returned, with the
    same dimensions as the variable, apart from the time dimension.

    Parameters
    ----------
    input_data
        Dataset with the input data. The input data dimensions determine the output
        data dimensions: the output data has the same dimensions as the input data,
        minus the priority dimensions defined in the priority_definition. From the
        priority dimensions, the different datasets for the filling are selected, so
        they vanish in the result.
    priority_definition
        Defines the priorities to select timeseries from the input data. Priorities
        are formed by a list of selections and are used "from left to right", where the
        first matching selection has the highest priority. Each selection has to specify
        values for all priority dimensions (so that exactly one timeseries is selected
        from the input data), but can also specify other dimensions. That way it is,
        e.g., possible to define a different priority for a specific country by listing
        it early (i.e. with high priority) before the more general rules which should
        be applied for all other countries.
        You can also specify the "entity" or "variable" in the selection, which will
        limit the rule to a specific entity or variable, respectively. For each
        DataArray in the input_data Dataset, the variable is its name, the entity is
        the value of the key `entity` in its attrs.
    strategy_definition
        Defines the filling strategies to be used when filling timeseries with other
        timeseries. Again, the priority is defined by a list of selections and
        corresponding strategies which are used "from left to right". Selections can use
        any dimension and don't have to apply to only one timeseries. For example, to
        define a default strategy which should be used for all timeseries unless
        something else is configured, configure an empty selection as the last
        (rightmost) entry.
        You can also specify the "entity" or "variable" in the selection, which will
        limit the rule to a specific entity or variable, respectively. For each
        DataArray in the input_data Dataset, the variable is its name, the entity is
        the value of the key `entity` in its attrs.
    progress_bar
        By default, show progress bars using the tqdm package during the
        operation. If None, don't show any progress bars. You can supply a class
        compatible to tqdm.tqdm's protocol if you want to customize the progress bar.

    Returns
    -------
        result. Dataset with the same entities and dimensions as input_data, but with
        following changes: the data is composed and filled according to the rules,
        the priority dimensions are reduced and not included in the result, and
        additional variables of the form "Processing of $variable" are added which
        describe the processing steps done for each timeseries.
    """
    result_das = {}
    input_data = input_data.pr.dequantify()

    if progress_bar is None:
        variable_iterator = input_data
    else:
        variable_iterator = progress_bar(input_data)

    priority_dimensions = priority_definition.priority_dimensions
    priority_definition.check_dimensions()

    strategy_definition.check_dimensions(input_data)

    for variable in variable_iterator:
        if progress_bar is not None:
            variable_iterator.set_postfix_str(str(variable))

        input_da = input_data[variable]
        entity = input_da.attrs["entity"]

        # all dimensions are either time, priority selection dimensions, or need to
        # be iterated over
        group_by_dimensions = tuple(
            dim
            for dim in input_da.dims
            if dim != "time" and dim not in priority_dimensions
        )

        (
            result_das[variable],
            result_das[f"Processing of {variable}"],
        ) = preallocate_result_arrays(
            input_da=input_da,
            group_by_dimensions=group_by_dimensions,
            priority_dimensions=priority_dimensions,
        )

        number_of_timeseries = math.prod(
            len(input_da[dim]) for dim in group_by_dimensions
        )

        if progress_bar is None:
            pbar = None
        else:
            pbar = progress_bar(total=number_of_timeseries, unit="ts", unit_scale=True)

        iterate_next_fixed_dimension(
            input_da=input_da,
            priority_definition=priority_definition.limit("entity", entity).limit(
                "variable", variable
            ),
            strategy_definition=strategy_definition.limit("entity", entity).limit(
                "variable", variable
            ),
            group_by_dimensions=group_by_dimensions,
            result_da=result_das[variable],
            result_processing_da=result_das[f"Processing of {variable}"],
            progress_bar=pbar,
        )
        if pbar is not None:
            pbar.close()

    result_ds = xr.Dataset(result_das).pr.quantify()
    # composing removes the priority dimensions, also remove the attrs describing
    # the priority dimensions
    result_ds.attrs = input_data.attrs.copy()
    for dim_key in priority_dimensions:
        if isinstance(dim_key, str) and "(" in dim_key:
            dim = dim_key.split("(", 1)[0][:-1]
        else:
            dim = dim_key
        if dim == "area":
            del result_ds.attrs["area"]
        elif dim == "category":
            del result_ds.attrs["cat"]
        elif dim == "scenario":
            del result_ds.attrs["scen"]
        elif dim not in ("provenance", "model", "source"):
            # remove from sec_cats if it is in sec_cats
            with contextlib.suppress(ValueError):
                result_ds.attrs["sec_cats"].remove(dim_key)
    return result_ds


def preallocate_result_arrays(
    *,
    group_by_dimensions: typing.Iterable[Hashable],
    input_da: xr.DataArray,
    priority_dimensions: typing.Iterable[Hashable],
) -> tuple[xr.DataArray, xr.DataArray]:
    """Create empty arrays in the right shape to hold the result."""
    result_da = xr.DataArray(
        data=np.nan,
        dims=["time", *group_by_dimensions],
        coords={
            dim: input_da.coords[dim]
            for dim in input_da.coords
            if dim not in priority_dimensions
        },
        attrs=input_da.attrs,
        name=input_da.name,
    )
    processing_result_da = xr.DataArray(
        data=np.empty(
            [len(input_da.coords[dim]) for dim in group_by_dimensions], dtype=object
        ),
        dims=group_by_dimensions,
        coords=[input_da.coords[dim] for dim in group_by_dimensions],
        attrs={
            "entity": f"Processing of {input_da.name}",
            "described_variable": input_da.name,
        },
    )
    return result_da, processing_result_da


def priority_coordinates_repr(
    *, fill_ts: xr.DataArray, priority_dimensions: list[Hashable]
) -> str:
    """Reduce the priority coordinates to a short string representation."""
    priority_coordinates: dict[str, str] = {
        str(k): fill_ts[k].item() for k in priority_dimensions
    }
    if len(priority_coordinates) == 1:
        # only one priority dimension, just output the value because it is clear what is
        # meant
        return repr(next(iter(priority_coordinates.values())))
    return repr(priority_coordinates)


def iterate_next_fixed_dimension(
    *,
    input_da: xr.DataArray,
    priority_definition: _models.PriorityDefinition,
    strategy_definition: _models.StrategyDefinition,
    group_by_dimensions: tuple[Hashable, ...],
    result_da: xr.DataArray,
    result_processing_da: xr.DataArray,
    progress_bar: tqdm.tqdm | None,
) -> None:
    """Recursively iterate over dimensions in group_by_dimensions.

    If there is only one dimension left, actually compute results and store
    them in the result_da and the result_processing_da. Otherwise, iterate over one
    dimension and recursively call iterate_next_fixed_dimension again.
    """
    my_dim = group_by_dimensions[0]
    new_group_by_dimensions = group_by_dimensions[1:]
    for val_array in input_da[my_dim]:
        val = val_array.item()
        limited_strategy_definition = strategy_definition.limit(dim=my_dim, value=val)
        limited_priority_definition = priority_definition.limit(dim=my_dim, value=val)
        if new_group_by_dimensions:
            # have to iterate further until all dimensions are consumed
            iterate_next_fixed_dimension(
                input_da=input_da.loc[{my_dim: val}],
                priority_definition=limited_priority_definition,
                strategy_definition=limited_strategy_definition,
                group_by_dimensions=new_group_by_dimensions,
                result_da=result_da.loc[{my_dim: val}],
                result_processing_da=result_processing_da.loc[{my_dim: val}],
                progress_bar=progress_bar,
            )
        else:
            # Result exclusions are handled here (per definition, we don't do any
            # processing on result exclusions) but input data exclusions are handled
            # in compose_timeseries (per definition, we skip to the next source when
            # a source is skipped due to input data exclusions).
            if not limited_priority_definition.excludes_result(
                result_da.loc[{my_dim: val}]
            ):
                # actually compute results
                (
                    result_da.loc[{my_dim: val}],
                    result_processing_da.loc[{my_dim: val}],
                ) = compose_timeseries(
                    input_data=input_da.loc[{my_dim: val}],
                    priority_definition=limited_priority_definition,
                    strategy_definition=limited_strategy_definition,
                )
            if progress_bar is not None:
                progress_bar.update()


def compose_timeseries(
    *,
    input_data: xr.DataArray,
    priority_definition: _models.PriorityDefinition,
    strategy_definition: _models.StrategyDefinition,
) -> tuple[xr.DataArray, primap2._data_format.TimeseriesProcessingDescription]:
    """
    Compute a single timeseries from given input data, priorities, and strategies.

    Parameters
    ----------
    input_data
        The input data which has dimensions time plus the priority dimensions. The
        fixed coordinates are supplied as zero-dimensional coordinates.
    priority_definition
        The definition of priorities within input_data. Each priority selects a single
        timeseries (i.e. an array with only time as the dimension), so has to specify
        values for all priority dimensions.
    strategy_definition
        The definition of strategies for timeseries in input_data.

    Returns
    -------
        result_ts, processing_description. In result_ts is the numerical result, with
        the time as the only dimension.
        processing_description is the representation of the processing steps taken to
        derive the result.
    """
    context_logger = logger.bind(
        fixed_coordinates={k: v for k, v in input_data.coords.items() if v.shape == ()},
        priority_coordinates={
            k: list(v.data) for k, v in input_data.coords.items() if v.shape != ()
        },
        priorities=priority_definition.priorities,
        strategies=strategy_definition.strategies,
    )

    result_ts: None | xr.DataArray = None
    processing_steps_descriptions = []
    for selector in priority_definition.priorities:
        try:
            fill_ts = input_data.loc[selector]
        except KeyError:
            context_logger.debug(f"{selector=} matched no input_data, skipping.")
            continue

        fill_ts_repr = priority_coordinates_repr(
            fill_ts=fill_ts,
            priority_dimensions=priority_definition.priority_dimensions,
        )
        # remove priority dimension information, it messes with automatic alignment
        # in computations. The corresponding information is now in fill_ts_repr.
        fill_ts_no_prio_dims = fill_ts.drop_vars(
            priority_definition.priority_dimensions
        )

        if result_ts is None:
            result_ts = xr.full_like(fill_ts_no_prio_dims, np.nan)

        if priority_definition.excludes_input(fill_ts):
            processing_steps_descriptions.append(
                primap2.ProcessingStepDescription(
                    time="all",
                    description=f"{fill_ts_repr} is excluded from processing, skipped",
                    function="compose_timeseries",
                    source=fill_ts_repr,
                )
            )
            continue
        if fill_ts.isnull().all():
            processing_steps_descriptions.append(
                primap2.ProcessingStepDescription(
                    time="all",
                    description=f"{fill_ts_repr} is fully NaN, skipped",
                    function="compose_timeseries",
                    source=fill_ts_repr,
                )
            )
            continue

        context_logger.debug(f"Filling with {fill_ts_repr} now.")

        for strategy in strategy_definition.find_strategies(fill_ts):
            try:
                result_ts, descriptions = strategy.fill(
                    ts=result_ts,
                    fill_ts=fill_ts_no_prio_dims,
                    fill_ts_repr=fill_ts_repr,
                )
                processing_steps_descriptions += descriptions
                break
            except primap2.csg.StrategyUnableToProcess:
                processing_steps_descriptions.append(
                    primap2.ProcessingStepDescription(
                        time="all",
                        description=f"strategy {strategy.type} unable to process "
                        f"{fill_ts_repr}, skipping to next strategy",
                        function="compose_timeseries",
                        source=fill_ts_repr,
                    )
                )
                # next strategy
                continue
        else:
            # only executed if there was no `break` in the for loop, i.e. if no strategy
            # was applicable and worked
            raise ValueError(
                f"No configured strategy was able to process "
                f"\n{input_data.coords}\n{input_data.attrs}\n{strategy_definition=}"
            )

        if not result_ts.isnull().any():
            context_logger.debug("No NaNs remaining, skipping the rest of the sources.")
            break

    if result_ts is None:
        raise ValueError(
            f"No priority selector matched for "
            f"\n{input_data.coords}\n{input_data.attrs}\n{priority_definition=}"
        )

    return result_ts, primap2._data_format.TimeseriesProcessingDescription(
        steps=processing_steps_descriptions
    )



"""Tests for the interchange format."""

import csv

import pandas as pd
import pytest
import xarray as xr

from primap2 import pm2io

from . import utils


def test_round_trip(any_ds: xr.Dataset, tmp_path):
    path = tmp_path / "if"
    pm2io.write_interchange_format(path, any_ds.pr.to_interchange_format())
    with path.with_suffix(".yaml").open() as fd:
        print(fd.read())
    actual = pm2io.from_interchange_format(pm2io.read_interchange_format(path))
    # we expect that Processing information is lost here
    expected = any_ds
    to_remove = []
    for var in expected:
        if (
            isinstance(var, str)
            and var.startswith("Processing of ")
            and "described_variable" in expected[var].attrs
        ):
            to_remove.append(var)
    for var in to_remove:
        del expected[var]
    utils.assert_ds_aligned_equal(any_ds, actual)


def test_missing_file(minimal_ds, tmp_path):
    path = tmp_path / "if"
    pm2io.write_interchange_format(path, minimal_ds.pr.to_interchange_format())
    with path.with_suffix(".yaml").open() as fd:
        content = fd.readlines()
    with path.with_suffix(".yaml").open("w") as fd:
        for line in content:
            if "data_file" in line:
                continue
            fd.write(line)

    # first test automatic discovery
    actual = pm2io.from_interchange_format(pm2io.read_interchange_format(path))
    utils.assert_ds_aligned_equal(minimal_ds, actual)

    # now test without csv file
    path.with_suffix(".csv").unlink()
    with pytest.raises(FileNotFoundError, match="Data file not found at"):
        pm2io.read_interchange_format(path)


def test_inharmonic_units(minimal_ds, tmp_path):
    path = tmp_path / "if"
    pm2io.write_interchange_format(path, minimal_ds.pr.to_interchange_format())
    df = pd.read_csv(path.with_suffix(".csv"))
    df.loc[3, "unit"] = "m"
    df.to_csv(path.with_suffix(".csv"), index=False, quoting=csv.QUOTE_NONNUMERIC)

    with pytest.raises(ValueError, match="More than one unit"):
        pm2io.from_interchange_format(pm2io.read_interchange_format(path))



#!/usr/bin/env python
"""Tests for _fill_combine.py

We only test the features regarding the buggy treatment of additional (non-indexed)
coordinates here. All core functionality is assumed to be sufficiently tested in xarray
"""

import numpy as np


def test_fillna_ds_coord_present(minimal_ds):
    # add additional coordinate
    country_names = ["Colombia", "Argentina", "Mexico", "Bolivia"]
    full_ds = minimal_ds.assign_coords(country_name=("area (ISO3)", country_names))

    sel = {"area (ISO3)": ["COL", "MEX"]}
    sel_ds = full_ds.pr.loc[sel]
    nan_ds = full_ds.copy()
    nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] = (
        nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] * np.nan
    )

    result_ds = nan_ds.pr.fillna(sel_ds)

    # assert_ds_aligned_equal(result_ds, full_ds)
    # above fails because data type of country_names differs

    # check that the additional coord is present in result
    assert "country_name" in list(result_ds.coords)
    # check that the mapping of country names to country codes is intact
    # (meaning that the additional coordinate is aligned correctly)
    for country in full_ds.coords["area (ISO3)"].values:
        assert (
            result_ds.coords["country_name"].loc[{"area (ISO3)": country}]
            == full_ds.coords["country_name"].loc[{"area (ISO3)": country}]
        )


def test_fillna_da_coord_present(minimal_ds):
    # add additional coordinate
    country_names = ["Colombia", "Argentina", "Mexico", "Bolivia"]
    full_ds = minimal_ds.assign_coords(country_name=("area (ISO3)", country_names))

    sel = {"area (ISO3)": ["COL", "MEX"]}
    sel_ds = full_ds.pr.loc[sel]
    nan_ds = full_ds.copy()
    nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] = (
        nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] * np.nan
    )

    result_da = nan_ds["CO2"].pr.fillna(sel_ds["CO2"])

    # check that the additional coord is present in result
    assert "country_name" in list(result_da.coords)
    # check that the mapping of country names to country codes is intact
    # (meaning that the additional coordinate is aligned correctly)
    for country in full_ds.coords["area (ISO3)"].values:
        assert (
            result_da.coords["country_name"].loc[{"area (ISO3)": country}]
            == full_ds.coords["country_name"].loc[{"area (ISO3)": country}]
        )


def test_combine_first_ds_coord_present(minimal_ds):
    # add additional coordinate
    country_names = ["Colombia", "Argentina", "Mexico", "Bolivia"]
    full_ds = minimal_ds.assign_coords(country_name=("area (ISO3)", country_names))

    sel = {"area (ISO3)": ["COL", "MEX"]}
    sel_ds = full_ds.pr.loc[sel]
    nan_ds = full_ds.pr.loc[{"area (ISO3)": ["ARG", "COL"]}]
    nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] = (
        nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] * np.nan
    )

    result_ds = nan_ds.pr.combine_first(sel_ds)
    compare_ds = full_ds.pr.loc[{"area (ISO3)": ["ARG", "COL", "MEX"]}]

    # check that the additional coord is present in result
    assert "country_name" in list(result_ds.coords)
    # check that the mapping of country names to country codes is intact
    # (meaning that the additional coordinate is aligned correctly)
    for country in compare_ds.coords["area (ISO3)"].values:
        assert (
            result_ds.coords["country_name"].loc[{"area (ISO3)": country}]
            == compare_ds.coords["country_name"].loc[{"area (ISO3)": country}]
        )


def test_combine_first_da_coord_present(minimal_ds):
    # add additional coordinate
    country_names = ["Colombia", "Argentina", "Mexico", "Bolivia"]
    full_ds = minimal_ds.assign_coords(country_name=("area (ISO3)", country_names))

    sel = {"area (ISO3)": ["COL", "MEX"]}
    sel_ds = full_ds.pr.loc[sel]
    nan_ds = full_ds.pr.loc[{"area (ISO3)": ["ARG", "COL"]}]
    nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] = (
        nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] * np.nan
    )

    result_da = nan_ds["CO2"].pr.combine_first(sel_ds["CO2"])
    compare_da = full_ds["CO2"].pr.loc[{"area (ISO3)": ["ARG", "COL", "MEX"]}]

    # check that the additional coord is present in result
    assert "country_name" in list(result_da.coords)
    # check that the mapping of country names to country codes is intact
    # (meaning that the additional coordinate is aligned correctly)
    for country in compare_da.coords["area (ISO3)"].values:
        assert (
            result_da.coords["country_name"].loc[{"area (ISO3)": country}]
            == compare_da.coords["country_name"].loc[{"area (ISO3)": country}]
        )


# tests to check if xarray bug persists
def test_fillna_ds_xr_fail(minimal_ds):
    # add additional coordinate
    country_names = ["Colombia", "Argentina", "Mexico", "Bolivia"]
    full_ds = minimal_ds.assign_coords(country_name=("area (ISO3)", country_names))

    sel = {"area (ISO3)": ["COL", "MEX"]}
    sel_ds = full_ds.pr.loc[sel]
    nan_ds = full_ds.copy()
    nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] = (
        nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] * np.nan
    )

    result_ds = nan_ds.fillna(sel_ds)

    assert "country_name" not in list(result_ds.coords)


def test_combine_first_ds_xr_fail(minimal_ds):
    # add additional coordinate
    country_names = ["Colombia", "Argentina", "Mexico", "Bolivia"]
    full_ds = minimal_ds.assign_coords(country_name=("area (ISO3)", country_names))

    sel = {"area (ISO3)": ["COL", "MEX"]}
    sel_ds = full_ds.pr.loc[sel]
    nan_ds = full_ds.pr.loc[{"area (ISO3)": ["ARG", "COL"]}]
    nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] = (
        nan_ds["CO2"].pr.loc[{"area (ISO3)": "COL"}] * np.nan
    )

    result_ds = nan_ds.combine_first(sel_ds)

    assert "country_name" not in list(result_ds.coords)



import pytest

import primap2.csg
from primap2 import Not
from primap2.csg._models import match_selector
from primap2.tests.csg.utils import get_single_ts


def test_match_selector():
    da = get_single_ts(coords={"source": "A", "category": "1.A"}, entity="SF6")

    assert match_selector(selector={"source": "A"}, ts=da)
    assert not match_selector(selector={"source": "B"}, ts=da)
    assert match_selector(selector={"source": "A", "category": "1.A"}, ts=da)
    assert match_selector(selector={"source": "A", "category": ["1.A", "1.B"]}, ts=da)
    assert not match_selector(selector={"source": "A", "category": "1"}, ts=da)
    assert not match_selector(selector={"source": "A", "category": ["1", "2"]}, ts=da)

    assert match_selector(selector={"source": "A", "entity": "SF6"}, ts=da)
    assert match_selector(selector={"source": "A", "entity": ["SF6", "CO2"]}, ts=da)
    assert not match_selector(selector={"source": "A", "entity": "CO2"}, ts=da)
    assert match_selector(selector={"source": "A", "variable": "SF6"}, ts=da)

    da = get_single_ts(
        coords={"source": "A", "category": "1.A"}, entity="SF6", gwp_context="AR6GWP100"
    )
    assert match_selector(selector={"source": "A", "entity": "SF6"}, ts=da)
    assert match_selector(selector={"source": "A", "entity": ["SF6", "CO2"]}, ts=da)
    assert not match_selector(selector={"source": "A", "entity": "CO2"}, ts=da)
    assert match_selector(
        selector={"source": "A", "variable": "SF6 (AR6GWP100)"}, ts=da
    )
    assert not match_selector(selector={"source": "A", "variable": "SF6"}, ts=da)


def test_selector_match_single_dim():
    assert primap2.csg.StrategyDefinition.match_single_dim(
        selector={"source": "A"}, dim="source", value="A"
    )
    assert not primap2.csg.StrategyDefinition.match_single_dim(
        selector={"source": "B"}, dim="source", value="A"
    )
    assert primap2.csg.StrategyDefinition.match_single_dim(
        selector={"source": ["A", "B"], "category": "1.A"}, dim="source", value="A"
    )
    assert primap2.csg.StrategyDefinition.match_single_dim(
        selector={"source": "A", "category": "1"}, dim="other", value="any"
    )


def test_strategy_definition():
    da = get_single_ts(
        coords={"source": "A", "category": "1.A"}, entity="SF6", gwp_context="AR6GWP100"
    )

    assert (
        primap2.csg.StrategyDefinition(
            [({"source": "A", "category": "1"}, 1), ({"source": "A"}, 2)]
        ).find_strategy(da)
        == 2
    )
    assert (
        primap2.csg.StrategyDefinition(
            [
                ({"source": "A", "category": "1.A", "entity": ["SF6", "CO2"]}, 1),
                ({"source": "A"}, 2),
            ]
        ).find_strategy(da)
        == 1
    )
    assert (
        primap2.csg.StrategyDefinition(
            [
                ({"source": "A", "category": "1.A", "entity": ["CH4", "CO2"]}, 1),
                ({"source": "A"}, 2),
            ]
        ).find_strategy(da)
        == 2
    )
    assert (
        primap2.csg.StrategyDefinition(
            [
                (
                    {
                        "source": "A",
                        "category": "1.A",
                        "variable": ["SF6 (AR6GWP100)", "CO2"],
                    },
                    1,
                ),
                ({"source": "A"}, 2),
            ]
        ).find_strategy(da)
        == 1
    )
    assert (
        primap2.csg.StrategyDefinition(
            [
                (
                    {
                        "source": "A",
                        "category": "1.A",
                        "variable": ["SF6 (AR4GWP100)", "CO2"],
                    },
                    1,
                ),
                ({"source": "A"}, 2),
            ]
        ).find_strategy(da)
        == 2
    )
    assert (
        primap2.csg.StrategyDefinition(
            [
                ({"source": "A", "category": "1"}, 1),
                ({"source": "A", "category": "1.A"}, 2),
            ]
        ).find_strategy(da)
        == 2
    )
    with pytest.raises(KeyError):
        primap2.csg.StrategyDefinition(
            [
                ({"source": "A", "category": "1"}, 1),
                ({"source": "A", "category": "1.B"}, 2),
                ({"source": "B", "category": "1.B"}, 3),
            ]
        ).find_strategy(da)


def test_strategy_definition_limit():
    assert primap2.csg.StrategyDefinition(
        [({"entity": "A", "source": "S"}, 1), ({"source": "T"}, 2)]
    ).limit("entity", "A").strategies == [({"source": "S"}, 1), ({"source": "T"}, 2)]
    assert primap2.csg.StrategyDefinition(
        [({"entity": "A", "source": "S"}, 1), ({"source": "T"}, 2)]
    ).limit("entity", "B").strategies == [({"source": "T"}, 2)]


def test_strategy_definition_check_dimensions(minimal_ds):
    primap2.csg.StrategyDefinition(
        [({"entity": "CO2", "source": "RAND2020"}, 1)]
    ).check_dimensions(minimal_ds)
    primap2.csg.StrategyDefinition(
        [({"variable": "CO2", "source": "RAND2020"}, 1)]
    ).check_dimensions(minimal_ds)
    with pytest.raises(ValueError):
        primap2.csg.StrategyDefinition(
            [({"entity": "CO2", "scenario": "highpop"})]
        ).check_dimensions(minimal_ds)


def test_priority_limit():
    pd = primap2.csg.PriorityDefinition(
        priority_dimensions=["a", "b"],
        priorities=[
            {
                "a": "1",
                "b": "2",
                "c": "3",
                "d": ["4", "5"],
                "e": Not("6"),
                "f": Not(["7", "8"]),
            },
            {"a": "2", "b": "3"},
        ],
        exclude_result=[{"c": "4"}],
        exclude_input=[{"a": "1", "c": "3", "d": "4"}],
    )
    assert pd.limit("g", "3") == pd
    assert pd.limit("c", "3").priorities == [
        {"a": "1", "b": "2", "d": ["4", "5"], "e": Not("6"), "f": Not(["7", "8"])},
        {"a": "2", "b": "3"},
    ]
    assert pd.limit("c", "4").priorities == [{"a": "2", "b": "3"}]
    assert pd.limit("d", "4").priorities == [
        {"a": "1", "b": "2", "c": "3", "e": Not("6"), "f": Not(["7", "8"])},
        {"a": "2", "b": "3"},
    ]
    assert pd.limit("d", "5") == pd.limit("d", "4")
    assert pd.limit("d", "6").priorities == [{"a": "2", "b": "3"}]
    assert pd.limit("e", "7").priorities == [
        {"a": "1", "b": "2", "c": "3", "d": ["4", "5"], "f": Not(["7", "8"])},
        {"a": "2", "b": "3"},
    ]
    assert pd.limit("e", "6").priorities == [{"a": "2", "b": "3"}]
    assert pd.limit("f", "6").priorities == [
        {"a": "1", "b": "2", "c": "3", "d": ["4", "5"], "e": Not("6")},
        {"a": "2", "b": "3"},
    ]
    assert pd.limit("f", "7").priorities == [{"a": "2", "b": "3"}]
    assert pd.limit("f", "7") == pd.limit("f", "8")


def test_priority_exclude_result():
    pd = primap2.csg.PriorityDefinition(
        priority_dimensions=["a"],
        priorities=[
            {"a": "1"},
            {"a": "2"},
        ],
        exclude_result=[{"b": "3", "c": ["4", "5"]}, {"c": "6"}],
    )
    pd.check_dimensions()

    assert pd.excludes_result(get_single_ts(coords={"a": "1", "b": "3", "c": "4"}))
    assert pd.excludes_result(get_single_ts(coords={"a": "1", "b": "3", "c": "5"}))
    assert pd.excludes_result(get_single_ts(coords={"a": "1", "b": "3", "c": "6"}))
    assert pd.excludes_result(get_single_ts(coords={"a": "1", "b": "4", "c": "6"}))
    assert not pd.excludes_result(get_single_ts(coords={"a": "1", "b": "3", "c": "7"}))
    assert not pd.excludes_result(get_single_ts(coords={"a": "1", "b": "4", "c": "4"}))


def test_priority_exclude_input():
    pd = primap2.csg.PriorityDefinition(
        priority_dimensions=["a"],
        priorities=[
            {"a": "1"},
            {"a": "2"},
        ],
        exclude_input=[
            {"a": "1", "b": "3", "c": ["4", "5"]},
            {"c": "6"},
        ],
    )
    pd.check_dimensions()

    assert pd.excludes_input(get_single_ts(coords={"a": "1", "b": "3", "c": "4"}))
    assert pd.excludes_input(get_single_ts(coords={"a": "1", "b": "3", "c": "5"}))
    assert pd.excludes_input(get_single_ts(coords={"a": "1", "b": "3", "c": "6"}))
    assert pd.excludes_input(get_single_ts(coords={"a": "1", "b": "4", "c": "6"}))
    assert not pd.excludes_input(get_single_ts(coords={"a": "2", "b": "3", "c": "4"}))
    assert not pd.excludes_input(get_single_ts(coords={"a": "1", "b": "3", "c": "7"}))
    assert not pd.excludes_input(get_single_ts(coords={"a": "1", "b": "4", "c": "4"}))


def test_priority_check():
    primap2.csg.PriorityDefinition(
        priority_dimensions=["a", "b"],
        priorities=[
            {"a": "1", "b": "2", "c": "3", "d": ["4", "5"]},
            {"a": "2", "b": "3"},
        ],
        exclude_result=[{"c": "5"}, {"d": ["6", "7"]}],
    ).check_dimensions()

    with pytest.raises(ValueError):
        primap2.csg.PriorityDefinition(
            priority_dimensions=["a", "b"],
            priorities=[
                {"a": "1", "b": "2", "c": "3", "d": ["4", "5"]},
                {"a": "2"},
            ],
        ).check_dimensions()

    with pytest.raises(ValueError):
        primap2.csg.PriorityDefinition(
            priority_dimensions=["a", "b"],
            priorities=[
                {"a": "1", "b": "2", "c": "3", "d": ["4", "5"]},
                {"a": "2", "b": ["2", "3"]},
            ],
        ).check_dimensions()

    with pytest.raises(ValueError):
        primap2.csg.PriorityDefinition(
            priority_dimensions=["a", "b"],
            priorities=[
                {"a": "1", "b": "2", "c": "3", "d": ["4", "5"]},
                {"a": "2", "b": "3"},
            ],
            exclude_result=[{"c": "5"}, {"d": ["6", "7"], "a": "3"}],
        ).check_dimensions()
